from __future__ import annotations

from importlib import import_module
from typing import Literal, Optional, Type, Union, cast

from ...model import Model
from ...models import Package, Resource


def convert_metadata(
    path: str,
    *,
    type: IType,
    format: Optional[str] = None,
    source: Optional[INotation] = None,
    target: Optional[INotation] = None,
) -> Model:
    """Convert metadata from one notation to another."""

    # Source model
    Source = Resource if type == "resource" else Package
    if source:
        module = import_module(f"dplib.plugins.{source}.models")
        Source: Type[Model] = getattr(module, f"{source.capitalize()}{type.capitalize()}")
    model = Source.from_path(path, format=format)
    if source:
        model = cast(Model, model.to_dp())  # type: ignore

    # Target model
    if target:
        module = import_module(f"dplib.plugins.{target}.models")
        Target: Type[Model] = getattr(module, f"{target.capitalize()}{type.capitalize()}")
        model = cast(Model, Target.from_dp(model))  # type: ignore

    return model


IType = Union[Literal["package"], Literal["resource"]]
INotation = Union[Literal["ckan"], Literal["dcat"], Literal["github"], Literal["zenodo"]]



from __future__ import annotations

from typing import List, Optional

from rdflib import BNode, Graph

from dplib.helpers.resource import slugify_name
from dplib.model import Model
from dplib.models import License, Resource

from . import dumpers, loaders
from . import namespaces as ns
from .types import ISubject


class DcatResource(Model):
    access_url: Optional[str] = None
    byte_size: Optional[int] = None
    conforms_to: List[str] = []
    description: Optional[str] = None
    download_url: Optional[str] = None
    issued: Optional[str] = None
    languages: List[str] = []
    license: Optional[str] = None
    media_type: Optional[str] = None
    modified: Optional[str] = None
    pages: List[str] = []
    title: Optional[str] = None

    # Converters

    def to_graph(self, g: Graph, *, id: BNode):
        # Access URL
        if self.access_url:
            dumpers.node(g, self.access_url, subject=id, predicate=ns.ACCESS_URL)

        # Byte size
        if self.byte_size:
            dumpers.node(g, self.byte_size, subject=id, predicate=ns.BYTE_SIZE)

        # Conforms to
        for conforms_to in self.conforms_to:
            dumpers.node(g, conforms_to, subject=id, predicate=ns.COMFORMS_TO)

        # Description
        if self.description:
            dumpers.node(g, self.description, subject=id, predicate=ns.DESCRIPTION)

        # Download URL
        if self.download_url:
            dumpers.node(g, self.download_url, subject=id, predicate=ns.DOWNLOAD_URL)

        # Issued
        if self.issued:
            dumpers.node(g, self.issued, subject=id, predicate=ns.ISSUED)

        # Languages
        for language in self.languages:
            dumpers.node(g, language, subject=id, predicate=ns.LANGUAGE)

        # License
        if self.license:
            dumpers.node(g, self.license, subject=id, predicate=ns.LICENSE)

        # Media type
        if self.media_type:
            dumpers.node(g, self.media_type, subject=id, predicate=ns.MEDIA_TYPE)

        # Modified
        if self.modified:
            dumpers.node(g, self.modified, subject=id, predicate=ns.MODIFIED)

        # Pages
        for page in self.pages:
            dumpers.node(g, page, subject=id, predicate=ns.PAGE)

        # Title
        if self.title:
            dumpers.node(g, self.title, subject=id, predicate=ns.TITLE)

        return g

    @classmethod
    def from_graph(cls, g: Graph, *, id: ISubject) -> DcatResource:
        resource = DcatResource()

        # Access URL
        access_url = loaders.string(g, subject=id, predicate=ns.ACCESS_URL)
        if access_url:
            resource.access_url = access_url

        # Byte size
        byte_size = loaders.integer(g, subject=id, predicate=ns.BYTE_SIZE)
        if byte_size:
            resource.byte_size = byte_size

        # Conforms to
        conforms_to = loaders.strings(g, subject=id, predicate=ns.COMFORMS_TO)
        if conforms_to:
            resource.conforms_to = conforms_to

        # Description
        description = loaders.string(g, subject=id, predicate=ns.DESCRIPTION)
        if description:
            resource.description = description

        # Download URL
        download_url = loaders.string(g, subject=id, predicate=ns.DOWNLOAD_URL)
        if download_url:
            resource.download_url = download_url

        # Issued
        issued = loaders.string(g, subject=id, predicate=ns.ISSUED)
        if issued:
            resource.issued = issued

        # Languages
        languages = loaders.strings(g, subject=id, predicate=ns.LANGUAGE)
        if languages:
            resource.languages = languages

        # License
        license = loaders.string(g, subject=id, predicate=ns.LICENSE)
        if license:
            resource.license = license

        # Media type
        media_type = loaders.string(g, subject=id, predicate=ns.MEDIA_TYPE)
        if media_type:
            resource.media_type = media_type

        # Modified
        modified = loaders.string(g, subject=id, predicate=ns.MODIFIED)
        if modified:
            resource.modified = modified

        # Pages
        pages = loaders.strings(g, subject=id, predicate=ns.PAGE)
        if pages:
            resource.pages = pages

        # Title
        title = loaders.string(g, subject=id, predicate=ns.TITLE)
        if title:
            resource.title = title

        return resource

    def to_dp(self) -> Optional[Resource]:
        if not self.download_url:
            return
        resource = Resource(path=self.download_url, name=slugify_name(self.download_url))

        # Title
        if self.title:
            resource.title = self.title

        # Description
        if self.description:
            resource.description = self.description

        # Media type
        if self.media_type:
            resource.mediatype = self.media_type

        # Bytes
        if self.byte_size:
            resource.bytes = self.byte_size

        # Licenses
        if self.license:
            license = License(path=self.license)
            resource.licenses.append(license)

        return resource

    @classmethod
    def from_dp(cls, resource: Resource) -> Optional[DcatResource]:
        dcat = DcatResource()
        if not resource.path or not isinstance(resource.path, str):
            return

        # Download URL
        # TODO: improve logic -- use basepath and allow only urls
        if resource.path:
            dcat.download_url = resource.path

        # Title
        if resource.title:
            dcat.title = resource.title

        # Description
        if resource.description:
            dcat.description = resource.description

        # Media type
        if resource.mediatype:
            dcat.media_type = resource.mediatype

        # Bytes
        if resource.bytes:
            dcat.byte_size = resource.bytes

        # Licenses
        if resource.licenses:
            license = resource.licenses[0]
            if license.path:
                dcat.license = license.path

        return dcat



from __future__ import annotations

import pprint
import warnings
from functools import cached_property
from typing import Optional

from pydantic import BaseModel
from typing_extensions import Self

from . import types
from .error import Error
from .helpers.data import clean_data, dump_data, load_data
from .helpers.file import read_file, write_file
from .helpers.path import ensure_basepath, infer_format


class Model(BaseModel, extra="allow", validate_assignment=True):
    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return pprint.pformat(self.to_dict(), sort_dicts=False)

    @cached_property
    def custom(self) -> types.IDict:
        assert self.model_extra is not None
        return self.model_extra

    # Converters

    def to_path(self, path: str, *, format: Optional[str] = None):
        if not format:
            format = infer_format(path, raise_missing=True)
        text = self.to_text(format=format)
        write_file(path, text)

    @classmethod
    def from_path(
        cls, path: str, *, format: Optional[str] = None, basepath: Optional[str] = None
    ) -> Self:
        if not format:
            format = infer_format(path, raise_missing=True)
        path, basepath = ensure_basepath(path, basepath=basepath)
        text = read_file(path)
        if not text:
            raise Error(f"The file is empty: {path}")
        return cls.from_text(text, format=format, basepath=basepath)

    def to_text(self, *, format: str) -> str:
        data = self.to_dict()
        text = dump_data(data, format=format)
        return text

    @classmethod
    def from_text(cls, text: str, *, format: str, basepath: Optional[str] = None) -> Self:
        data = load_data(text, format=format)
        return cls.from_dict(data, basepath=basepath)

    def to_dict(self):
        data = self.model_dump(mode="json", exclude_none=True, exclude_defaults=True)
        clean_data(data)
        return data

    @classmethod
    def from_dict(cls, data: types.IDict, *, basepath: Optional[str] = None) -> Self:
        if basepath and cls.model_fields.get("basepath"):
            data["basepath"] = basepath
        return cls(**data)


# Although pydantic@2 moved all the model methods to the "model_" namespace
# the "schema" method is still in the root namespace
# https://github.com/pydantic/pydantic/issues/5165
warnings.filterwarnings(
    action="ignore",
    category=UserWarning,
    module=r"pydantic.*",
    message=r".*schema.*",
)



from __future__ import annotations

import os
from pathlib import Path
from typing import Optional, Tuple
from urllib.parse import urlparse

from ..error import Error


def infer_format(path: str, *, raise_missing: bool = False):
    format = Path(path).suffix[1:]
    if format == "yml":
        format = "yaml"
    elif format == "rdf":
        format = "xml"
    if not format and raise_missing:
        raise Error(f"Cannot infer format from path: {path}")
    return format


def infer_basepath(path: str):
    basepath = os.path.dirname(path)
    if basepath and not is_remote_path(basepath):
        if not os.path.abspath(basepath):
            basepath = os.path.relpath(basepath, start=os.getcwd())
    return basepath


def ensure_basepath(path: str, basepath: Optional[str] = None) -> Tuple[str, str]:
    if basepath:
        path = join_basepath(path, basepath)
    else:
        basepath = infer_basepath(path)
    return path, basepath


def join_basepath(path: str, basepath: Optional[str] = None) -> str:
    if not basepath:
        return path
    if is_remote_path(path):
        return path
    if is_remote_path(basepath):
        return f"{basepath}/{path}"
    return os.path.join(basepath, path)


def is_remote_path(path: str) -> bool:
    path = path[0] if path and isinstance(path, list) else path
    scheme = urlparse(path).scheme
    if not scheme:
        return False
    if path.lower().startswith(scheme + ":\\"):
        return False
    return True



from typing import Any

from rdflib import Graph, URIRef

from .helpers import create_node
from .types import ISubject


def id(g: Graph, identifier: str, *, predicate: URIRef, object: URIRef):
    subject = URIRef(identifier)
    g.add((subject, predicate, object))
    return subject


def node(g: Graph, value: Any, *, subject: ISubject, predicate: URIRef):
    object = create_node(value)
    g.add((subject, predicate, object))



from __future__ import annotations

from typing import Optional

from dplib.helpers.resource import slugify_name
from dplib.model import Model
from dplib.models import Resource


class ZenodoResource(Model):
    key: str
    id: Optional[str] = None
    checksum: Optional[str] = None
    ext: Optional[str] = None
    mimetype: Optional[str] = None
    size: Optional[int] = None

    # Converters

    def to_dp(self) -> Resource:
        resource = Resource(path=self.key, name=slugify_name(self.key))

        # Format
        if self.ext:
            resource.format = self.ext.lower()

        # Mediatype
        if self.mimetype:
            resource.mediatype = self.mimetype

        # Bytes
        if self.size:
            resource.bytes = self.size

        # Hash
        if self.checksum:
            resource.hash = self.checksum.replace("md5:", "")

        return resource

    @classmethod
    def from_dp(cls, resource: Resource) -> Optional[ZenodoResource]:
        if not resource.path or not isinstance(resource.path, str):
            return

        # Path
        zenodo = ZenodoResource(key=resource.path)

        # Format
        if resource.format:
            zenodo.ext = resource.format

        # Mediatype
        if resource.mediatype:
            zenodo.mimetype = resource.mediatype

        # Bytes
        if resource.bytes:
            zenodo.size = resource.bytes

        # Hash
        hash = resource.get_hash()
        if hash:
            if hash.type == "md5":
                zenodo.checksum = hash.long

        return zenodo



from __future__ import annotations

from typing import Literal, Optional

from dplib.helpers.resource import slugify_name
from dplib.model import Model
from dplib.models import Resource


class GithubResource(Model):
    name: str
    path: str
    type: Literal["file"] = "file"
    content: Optional[str] = None
    encoding: Optional[str] = None
    size: Optional[int] = None
    sha: Optional[str] = None
    html_url: Optional[str] = None
    download_url: Optional[str] = None

    # Converters

    def to_dp(self):
        resource = Resource(path=self.path, name=slugify_name(self.path))

        # Bytes
        if self.size:
            resource.bytes = self.size

        # Hash
        if self.sha:
            resource.hash = f"sha1:{self.sha}"

        return resource

    @classmethod
    def from_dp(cls, resource: Resource) -> Optional[GithubResource]:
        if not resource.path or not isinstance(resource.path, str):
            return

        # Path
        github = GithubResource(path=resource.path, name=resource.path)

        # Bytes
        if resource.bytes:
            github.size = resource.bytes

        # Hash
        hash = resource.get_hash()
        if hash:
            if hash.type == "sha1":
                github.sha = hash.value

        return github



from typing import List, Optional

from rdflib import Graph, Literal, URIRef

from .types import IStringNode, ISubject


def id(g: Graph, *, predicate: URIRef, object: URIRef) -> Optional[URIRef]:
    try:
        id = g.value(predicate=predicate, object=object)
        if isinstance(id, URIRef):
            return id
    except Exception:
        pass


def node(g: Graph, *, subject: ISubject, predicate: URIRef) -> Optional[IStringNode]:
    default_lang = "en"
    items = list(g.objects(subject, predicate))

    # Prefer the default language
    for item in items:
        if isinstance(item, Literal):
            if item.language and item.language == default_lang:
                return item

    # Otherwise, return the first item
    for item in items:
        if isinstance(item, (URIRef, Literal)):
            return item


def string(g: Graph, *, subject: ISubject, predicate: URIRef) -> Optional[str]:
    value = node(g, subject=subject, predicate=predicate)
    if value:
        return str(value)


def integer(g: Graph, *, subject: ISubject, predicate: URIRef) -> Optional[int]:
    value = node(g, subject=subject, predicate=predicate)
    if value:
        try:
            return int(value)
        except Exception:
            pass


def nodes(g: Graph, *, subject: ISubject, predicate: URIRef) -> List[IStringNode]:
    return [
        item
        for item in g.objects(subject, predicate)
        if isinstance(item, (URIRef, Literal))
    ]


def strings(g: Graph, *, subject: ISubject, predicate: URIRef) -> List[str]:
    return [str(item) for item in nodes(g, subject=subject, predicate=predicate)]



from pathlib import Path

import pytest
from pydantic import ValidationError

from dplib.models import Dialect, Resource, Schema


def test_resource_from_path():
    resource = Resource.from_path("data/resource.json")
    assert resource.name == "name"
    assert resource.path == "table.csv"


def test_resource_from_path_full():
    resource = Resource.from_path("data/resource-full.json")
    assert resource.name == "name"
    assert resource.type == "table"
    assert resource.path == "table.csv"
    assert resource.title == "title"
    assert resource.description == "description"
    assert resource.format == "csv"
    assert resource.mediatype == "text/csv"
    assert resource.encoding == "utf-8"
    assert resource.bytes == 100
    dialect = resource.get_dialect()
    assert dialect
    assert dialect.delimiter == ";"
    schema = resource.get_schema()
    assert schema
    assert len(schema.fields) == 2


def test_resource_from_text():
    text = '{"name": "name"}'
    resource = Resource.from_text(text, format="json")
    assert resource.name == "name"


def test_resource_from_dict():
    data = {"name": "name"}
    resource = Resource.from_dict(data)
    assert resource.name == "name"


def test_resource_from_dict_invalid():
    data = {"name": 1}
    with pytest.raises(ValidationError):
        Resource.from_dict(data)


def test_resource_set_proprty_invalid():
    resource = Resource()
    with pytest.raises(ValidationError):
        resource.name = 1  # type: ignore


@pytest.mark.vcr
def test_resource_profile():
    resource = Resource.from_path("data/resource-full.json")
    profile = resource.get_profile()
    assert profile
    assert profile.jsonSchema.get("title") == "Tabular Data Resource"


def test_resource_get_fullpath():
    resource = Resource.from_path("data/resource.json")
    fullpath = resource.get_fullpath()
    assert fullpath == str(Path("data/table.csv"))


def test_resource_get_fullpath_with_basepath():
    resource = Resource(path="table.csv", basepath="data")
    fullpath = resource.get_fullpath()
    assert fullpath == str(Path("data/table.csv"))


def test_resource_get_source():
    resource = Resource(data=[])
    source = resource.get_source()
    assert source == []


def test_resource_dialect_inline():
    resource = Resource(dialect=Dialect(delimiter=";"))
    dialect = resource.get_dialect()
    assert dialect
    assert dialect.delimiter == ";"


def test_resource_schema_inline():
    resource = Resource(schema=Schema.from_path("data/schema.json"))
    schema = resource.get_schema()
    assert schema
    assert len(schema.fields) == 2


def test_resource_hash():
    resource = Resource(hash="hash")
    hash = resource.get_hash()
    assert hash
    assert hash.type == "md5"
    assert hash.value == "hash"
    assert hash.short == "hash"
    assert hash.long == "md5:hash"


def test_resource_hash_non_default_type():
    resource = Resource(hash="sha256:hash")
    hash = resource.get_hash()
    assert hash
    assert hash.type == "sha256"
    assert hash.value == "hash"
    assert hash.short == "sha256:hash"
    assert hash.long == "sha256:hash"


def test_resource_dereference():
    resource = Resource.from_path("data/resource-full.json")
    resource.dereference()
    assert isinstance(resource.dialect, Dialect)
    assert resource.dialect.delimiter == ";"
    assert isinstance(resource.schema, Schema)
    assert len(resource.schema.fields) == 2



from __future__ import annotations

from typing import Any, Callable, List

import sqlalchemy as sa
from sqlalchemy.schema import Column, Constraint, Table

from dplib.model import Model
from dplib.models import ForeignKey, ForeignKeyReference, Schema

from . import settings
from .field import SqlField


class SqlSchema(Model, arbitrary_types_allowed=True):
    table: Table

    # Getters

    def get_field_names(self) -> List[str]:
        names: List[str] = []
        for column in self.table.columns:
            names.append(column.name)
        return names

    def get_field_types(self) -> List[Any]:
        types: List[Any] = []
        for column in self.table.columns:
            types.append(type(column.type))
        return types

    # Converters

    def to_dp(self, *, with_metadata: bool = False) -> Schema:
        schema = Schema()

        # Fields
        for column in self.table.columns:
            if with_metadata and column.name in settings.METADATA_IDENTIFIERS:
                continue
            field = SqlField(column=column).to_dp()
            schema.fields.append(field)

        # Primary key
        for constraint in self.table.constraints:
            if isinstance(constraint, sa.PrimaryKeyConstraint):
                for column in constraint.columns:
                    if with_metadata and column.name in settings.METADATA_IDENTIFIERS:
                        continue
                    schema.primaryKey.append(str(column.name))

        # Foreign keys
        for constraint in self.table.constraints:
            if isinstance(constraint, sa.ForeignKeyConstraint):
                resource = ""
                own_fields: List[str] = []
                foreign_fields: List[str] = []
                for element in constraint.elements:
                    own_fields.append(str(element.parent.name))
                    if element.column.table.name != self.table.name:
                        resource = str(element.column.table.name)
                    foreign_fields.append(str(element.column.name))
                ref = ForeignKeyReference(resource=resource, fields=foreign_fields)
                fk = ForeignKey(fields=own_fields, reference=ref)
                schema.foreignKeys.append(fk)

        return schema

    @classmethod
    def from_dp(
        cls,
        schema: Schema,
        *,
        table_name: str,
        dialect: str = settings.DEFAULT_DIALECT,
        with_metadata: bool = False,
    ) -> SqlSchema:
        columns: List[Column[Any]] = []
        constraints: List[Constraint] = []

        # Fields
        if with_metadata:
            columns.append(
                sa.Column(
                    settings.ROW_NUMBER_IDENTIFIER,
                    sa.Integer,
                    primary_key=True,
                    autoincrement=False,
                )
            )
            columns.append(sa.Column(settings.ROW_VALID_IDENTIFIER, sa.Boolean))
        for field in schema.fields:
            sql_field = SqlField.from_dp(field, table_name=table_name, dialect=dialect)
            columns.append(sql_field.column)

        # Primary key
        if schema.primaryKey:
            Class = sa.UniqueConstraint if with_metadata else sa.PrimaryKeyConstraint
            if not with_metadata:
                constraint = Class(*schema.primaryKey)
                constraints.append(constraint)

        # Foreign keys
        for fk in schema.foreignKeys:
            prefix: Callable[[str], str] = lambda field: ".".join([foreign_table, field])
            foreign_table = fk.reference.resource or table_name
            foreign_fields = list(map(prefix, fk.reference.fields))
            constraint = sa.ForeignKeyConstraint(fk.fields, foreign_fields)
            constraints.append(constraint)

        table = sa.Table(table_name, sa.MetaData(), *(columns + constraints))
        return SqlSchema(table=table)



from __future__ import annotations

from typing import List, Optional

from dplib.model import Model


class DataciteContributorAffiliation(Model):
    name: Optional[str] = None
    affiliationIdentifier: Optional[str] = None
    affiliationIdentifierScheme: Optional[str] = None


class DataciteContributor(Model):
    name: Optional[str] = None
    nameType: Optional[str] = None
    givenName: Optional[str] = None
    familyName: Optional[str] = None
    contributorType: Optional[str] = None
    affiliation: List[DataciteContributorAffiliation] = []



from __future__ import annotations

from typing import List, Union

from ... import types
from ...errors.metadata import MetadataError
from ...models import Dialect
from ..metadata.check import check_metadata


def check_dialect(dialect: Union[str, types.IDict, Dialect]) -> List[MetadataError]:
    if isinstance(dialect, Dialect):
        dialect = dialect.to_dict()
    return check_metadata(dialect, type="dialect")



from __future__ import annotations

from typer import Option

Source = Option(None, "--source", "-s", help="Source notation e.g. ckan")
Target = Option(None, "--target", "-t", help="Target notation e.g. dcat")



from pathlib import Path

import pytest

from dplib.error import Error
from dplib.models import Resource


def test_model_repr():
    resource = Resource.from_path("data/resource.json")
    assert str(resource) == "{'name': 'name', 'path': 'table.csv'}"


def test_model_custom_properties():
    data = {"name": "name", "extra": "value"}
    resource = Resource.from_dict(data)
    assert resource.name == "name"
    assert resource.custom["extra"] == "value"
    assert resource.to_dict() == data


def test_model_to_path(tmp_path: Path):
    path = str(tmp_path / "resource.json")
    resource = Resource.from_path("data/resource.json")
    resource.to_path(path)
    with open(path) as file:
        assert file.read() == '{"name": "name", "path": "table.csv"}'


def test_model_from_path_emtpy_file():
    with pytest.raises(Error) as excinfo:
        Resource.from_path("data/empty.json")
    error = excinfo.value
    assert str(error) == "The file is empty: data/empty.json"


def test_model_from_path_missing_file():
    with pytest.raises(Error) as excinfo:
        Resource.from_path("data/missing.json")
    error = excinfo.value
    assert str(error) == 'Cannot read file "data/missing.json"'


def test_model_from_path_not_supported_format():
    with pytest.raises(Error) as excinfo:
        Resource.from_path("data/text.txt")
    error = excinfo.value
    assert str(error) == "Cannot load data from text with format: txt"


def test_model_from_path_yaml():
    resource = Resource.from_path("data/resource.yaml")
    assert resource.name == "name"
    assert resource.path == "table.csv"


def test_model_to_path_yaml(tmp_path: Path):
    path = str(tmp_path / "resource.yaml")
    resource = Resource.from_path("data/resource.json")
    resource.to_path(path)
    with open(path) as file:
        assert file.read() == "name: name\npath: table.csv\n"



from __future__ import annotations

from typing import Any, Dict, List, Optional

import pydantic

from ...model import Model
from .constraints import Constraints
from .fieldType import FieldType


# TODO: consider getting back to discriminated unions
class Field(Model):
    name: Optional[str] = None
    type: FieldType = "any"
    title: Optional[str] = None
    description: Optional[str] = None
    format: Optional[str] = None
    missingValues: List[str] = [""]
    constraints: Constraints = pydantic.Field(default_factory=Constraints)

    # Array
    arrayItem: Optional[Dict[str, Any]] = None

    # Boolean
    trueValues: Optional[List[str]] = None
    falseValues: Optional[List[str]] = None

    # Integer/Number
    bareNumber: Optional[bool] = None
    groupChar: Optional[str] = None
    decimalChar: Optional[str] = None



from dplib.models import Field


def test_field_defaults():
    field = Field()
    assert field.type == "any"
    assert field.missingValues == [""]


def test_field_constraints():
    field = Field()
    field.constraints.minimum = 1
    assert field.constraints.minimum == 1
    assert field.to_dict() == {"constraints": {"minimum": 1}}



from __future__ import annotations

from typing import List, Optional

from dplib.model import Model
from dplib.models import License, Package

from .license import GithubLicense
from .owner import GithubOwner
from .resource import GithubResource

# References:
# - https://docs.github.com/en/free-pro-team@latest/rest/repos/repos?apiVersion=2022-11-28#get-a-repository


class GithubPackage(Model):
    resources: List[GithubResource] = []

    license: Optional[GithubLicense] = None
    owner: Optional[GithubOwner] = None

    private: bool = False
    name: Optional[str] = None
    full_name: Optional[str] = None
    html_url: Optional[str] = None
    description: Optional[str] = None
    pushed_at: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    topics: List[str] = []

    # Converters

    def to_dp(self):
        package = Package()

        # Title
        if self.name:
            package.title = self.name

        # Name
        if self.full_name:
            package.name = self.full_name

        # Description
        if self.description:
            package.description = self.description

        # Homepage
        if self.html_url:
            package.homepage = self.html_url

        # Created
        if self.created_at:
            package.created = self.created_at

        # License
        if self.license:
            license = License()
            if self.license.spdx_id:
                license.name = self.license.spdx_id
            if self.license.name:
                license.title = self.license.name
            if self.license.html_url:
                license.url = self.license.html_url
            package.licenses.append(license)

        # Keywords
        for topic in self.topics:
            package.keywords.append(topic)

        # Resources
        for item in self.resources:
            resource = item.to_dp()
            package.resources.append(resource)

        return package

    @classmethod
    def from_dp(cls, package: Package) -> GithubPackage:
        github = GithubPackage()

        # Title
        if package.title:
            github.name = package.title

        # Description
        if package.description:
            github.description = package.description

        # Licenses
        if package.licenses:
            license = package.licenses[0]
            github.license = GithubLicense()
            if license.name:
                github.license.spdx_id = license.name
            if license.title:
                github.license.name = license.title
            if license.path:
                github.license.html_url = license.path

        # Keywords
        for keyword in package.keywords:
            github.topics.append(keyword)

        # Resources
        for resource in package.resources:
            item = GithubResource.from_dp(resource)
            if item:
                github.resources.append(item)

        return github



from __future__ import annotations

from typing import Dict, List

import polars as pl

from dplib.model import Model
from dplib.models import Schema

from .field import PolarsField


class PolarsSchema(Model, arbitrary_types_allowed=True):
    df: pl.DataFrame

    # Getters

    def get_field_names(self) -> List[str]:
        return list(self.df.schema.keys())

    def get_field_types(self) -> List[pl.PolarsDataType]:
        return list(self.df.schema.values())

    # Converters

    def to_dp(self) -> Schema:
        schema = Schema()

        # Fields
        for name, dtype in self.df.schema.items():
            field = PolarsField(name=name, dtype=dtype).to_dp()
            schema.fields.append(field)

        return schema

    @classmethod
    def from_dp(cls, schema: Schema) -> PolarsSchema:
        columns: Dict[str, pl.PolarsDataType] = {}

        # Fields
        for field in schema.fields:
            polars_field = PolarsField.from_dp(field)
            columns[polars_field.name] = polars_field.dtype

        return PolarsSchema(df=pl.DataFrame(schema=columns))



from __future__ import annotations

from typing import Any, List, Optional

from rdflib import BNode, Graph, URIRef

from dplib.error import Error
from dplib.model import Model
from dplib.models import Package

from . import dumpers, loaders
from . import namespaces as ns
from .resource import DcatResource

# References:
# - https://www.w3.org/TR/vocab-dcat-2/
# - https://joinup.ec.europa.eu/asset/dcat_application_profile
# - https://github.com/ckan/ckanext-dcat/blob/master/ckanext/dcat/profiles.py


class DcatPackage(Model):
    identifier: Optional[str] = None
    distributions: List[DcatResource] = []

    accural_periodicity: Optional[str] = None
    alternate_identifiers: List[str] = []
    comforms_to: List[str] = []
    description: Optional[str] = None
    has_versions: List[str] = []
    homepage: Optional[str] = None
    issued: Optional[str] = None
    is_version_of: List[str] = []
    keywords: List[str] = []
    landing_page: Optional[str] = None
    languages: List[str] = []
    modified: Optional[str] = None
    pages: List[str] = []
    provenance: Optional[str] = None
    related_resources: List[str] = []
    samples: List[str] = []
    sources: List[str] = []
    themes: List[str] = []
    title: Optional[str] = None
    version: Optional[str] = None

    # Converters

    def to_text(self, *, format: str):
        g = self.to_graph()
        return g.serialize(format=format)

    @classmethod
    def from_text(cls, text: str, *, format: str, **kwargs: Any):
        g = Graph()
        g.parse(data=text, format=format)
        return cls.from_graph(g)

    def to_graph(self):
        g = Graph()
        for prefix, namespace in ns.BINDINGS.items():
            g.bind(prefix, namespace)

        # Identifier
        if not self.identifier:
            raise Error(f"Cannot dump DCAT package without identifier: {self}")
        id = dumpers.id(g, self.identifier, predicate=ns.TYPE, object=ns.DATASET)

        # Accural periodicity
        if self.accural_periodicity:
            dumpers.node(
                g, self.accural_periodicity, subject=id, predicate=ns.ACCURAL_PERIODICITY
            )

        # Alternate identifiers
        for identifier in self.alternate_identifiers:
            dumpers.node(g, identifier, subject=id, predicate=ns.ALTERNATE_IDENTIFIER)

        # Conforms to
        for conforms_to in self.comforms_to:
            dumpers.node(g, conforms_to, subject=id, predicate=ns.COMFORMS_TO)

        # Description
        if self.description:
            dumpers.node(g, self.description, subject=id, predicate=ns.DESCRIPTION)

        # Has versions
        for has_version in self.has_versions:
            dumpers.node(g, has_version, subject=id, predicate=ns.HAS_VERSION)

        # Homepage
        if self.homepage:
            dumpers.node(g, self.homepage, subject=id, predicate=ns.HOMEPAGE)

        # Issued
        if self.issued:
            dumpers.node(g, self.issued, subject=id, predicate=ns.ISSUED)

        # Is version of
        for is_version_of in self.is_version_of:
            dumpers.node(g, is_version_of, subject=id, predicate=ns.IS_VERSION_OF)

        # Keywords
        for keyword in self.keywords:
            dumpers.node(g, keyword, subject=id, predicate=ns.KEYWORD)

        # Landing page
        if self.landing_page:
            dumpers.node(g, self.landing_page, subject=id, predicate=ns.LANDING_PAGE)

        # Languages
        for language in self.languages:
            dumpers.node(g, language, subject=id, predicate=ns.LANGUAGE)

        # Modified
        if self.modified:
            dumpers.node(g, self.modified, subject=id, predicate=ns.MODIFIED)

        # Pages
        for page in self.pages:
            dumpers.node(g, page, subject=id, predicate=ns.PAGE)

        # Provenance
        if self.provenance:
            dumpers.node(g, self.provenance, subject=id, predicate=ns.PROVENANCE)

        # Related resources
        for related_resource in self.related_resources:
            dumpers.node(g, related_resource, subject=id, predicate=ns.RELATED_RESOURCE)

        # Samples
        for sample in self.samples:
            dumpers.node(g, sample, subject=id, predicate=ns.SAMPLE)

        # Sources
        for source in self.sources:
            dumpers.node(g, source, subject=id, predicate=ns.SOURCE)

        # Themes
        for theme in self.themes:
            dumpers.node(g, theme, subject=id, predicate=ns.THEME)

        # Title
        if self.title:
            dumpers.node(g, self.title, subject=id, predicate=ns.TITLE)

        # Version
        if self.version:
            dumpers.node(g, self.version, subject=id, predicate=ns.VERSION)

        # Distributions
        for distribution in self.distributions:
            distribution_id = BNode()
            g.add((id, ns.DISTRIBUTION, distribution_id))
            g.add((distribution_id, ns.TYPE, ns.DISTRIBUTION))
            distribution.to_graph(g, id=distribution_id)

        return g

    @classmethod
    def from_graph(cls, g: Graph):
        package = DcatPackage()

        # Identifier
        id = loaders.id(g, predicate=ns.TYPE, object=ns.DATASET)
        if not id:
            raise Error(f"Cannot load DCAT package without identifier: {g}")
        package.identifier = str(id)

        # Accural periodicity
        periodicity = loaders.string(g, subject=id, predicate=ns.ACCURAL_PERIODICITY)
        if periodicity:
            package.accural_periodicity = periodicity

        # Alternate identifiers
        identifiers = loaders.strings(g, subject=id, predicate=ns.ALTERNATE_IDENTIFIER)
        if identifiers:
            package.alternate_identifiers = identifiers

        # Conforms to
        conforms_to = loaders.strings(g, subject=id, predicate=ns.COMFORMS_TO)
        if conforms_to:
            package.comforms_to = conforms_to

        # Description
        description = loaders.string(g, subject=id, predicate=ns.DESCRIPTION)
        if description:
            package.description = description

        # Has versions
        has_versions = loaders.strings(g, subject=id, predicate=ns.HAS_VERSION)
        if has_versions:
            package.has_versions = has_versions

        # Homepage
        homepage = loaders.string(g, subject=id, predicate=ns.HOMEPAGE)
        if homepage:
            package.homepage = homepage

        # Issued
        issued = loaders.string(g, subject=id, predicate=ns.ISSUED)
        if issued:
            package.issued = issued

        # Is version of
        is_version_of = loaders.strings(g, subject=id, predicate=ns.IS_VERSION_OF)
        if is_version_of:
            package.is_version_of = is_version_of

        # Keywords
        keywords = loaders.strings(g, subject=id, predicate=ns.KEYWORD)
        if keywords:
            package.keywords = keywords

        # Landing page
        landing_page = loaders.string(g, subject=id, predicate=ns.LANDING_PAGE)
        if landing_page:
            package.landing_page = landing_page

        # Languages
        languages = loaders.strings(g, subject=id, predicate=ns.LANGUAGE)
        if languages:
            package.languages = languages

        # Modified
        modified = loaders.string(g, subject=id, predicate=ns.MODIFIED)
        if modified:
            package.modified = modified

        # Pages
        pages = loaders.strings(g, subject=id, predicate=ns.PAGE)
        if pages:
            package.pages = pages

        # Provenance
        provenance = loaders.string(g, subject=id, predicate=ns.PROVENANCE)
        if provenance:
            package.provenance = provenance

        # Related resources
        related_resources = loaders.strings(g, subject=id, predicate=ns.RELATED_RESOURCE)
        if related_resources:
            package.related_resources = related_resources

        # Samples
        samples = loaders.strings(g, subject=id, predicate=ns.SAMPLE)
        if samples:
            package.samples = samples

        # Sources
        sources = loaders.strings(g, subject=id, predicate=ns.SOURCE)
        if sources:
            package.sources = sources

        # Themes
        themes = loaders.strings(g, subject=id, predicate=ns.THEME)
        if themes:
            package.themes = themes

        # Title
        title = loaders.string(g, subject=id, predicate=ns.TITLE)
        if title:
            package.title = title

        # Version
        version = loaders.string(g, subject=id, predicate=ns.VERSION)
        if version:
            package.version = version

        # Distributions
        distributions = g.objects(subject=id, predicate=ns.DISTRIBUTION)
        for distribution in distributions:
            if isinstance(distribution, (URIRef, BNode)):
                resource = DcatResource.from_graph(g, id=distribution)
                package.distributions.append(resource)

        return package

    def to_dp(self):
        package = Package()

        # Id
        if self.identifier:
            package.id = self.identifier

        # Title
        if self.title:
            package.title = self.title

        # Description
        if self.description:
            package.description = self.description

        # Version
        if self.version:
            package.version = self.version

        # Homepage
        if self.homepage:
            package.homepage = self.homepage

        # Created
        if self.issued:
            if "T" in self.issued:
                package.created = self.issued

        # Keywords
        for keyword in self.keywords:
            package.keywords.append(keyword)

        # Resources
        for distribution in self.distributions:
            resource = distribution.to_dp()
            if resource:
                package.resources.append(resource)

        return package

    @classmethod
    def from_dp(cls, package: Package):
        dcat = DcatPackage()

        # Identifier
        if package.id:
            dcat.identifier = package.id

        # Title
        if package.title:
            dcat.title = package.title

        # Description
        if package.description:
            dcat.description = package.description

        # Version
        if package.version:
            dcat.version = package.version

        # Homepage
        if package.homepage:
            dcat.homepage = package.homepage

        # Issued
        if package.created:
            dcat.issued = package.created

        # Keywords
        for keyword in package.keywords:
            dcat.keywords.append(keyword)

        # Resources
        for resource in package.resources:
            distribution = DcatResource.from_dp(resource)
            if distribution:
                dcat.distributions.append(distribution)

        return dcat



from __future__ import annotations

from typing import Any, Dict

import pandas as pd

from dplib.model import Model
from dplib.models import Schema

from .field import PandasField


class PandasSchema(Model, arbitrary_types_allowed=True):
    df: pd.DataFrame

    # Converters

    def to_dp(self) -> Schema:
        schema = Schema()

        # Primary key
        for index, name in enumerate(self.df.index.names):  # type: ignore
            dtype = self.df.index.get_level_values(index).dtype  # type: ignore
            field = PandasField(name=name, dtype=dtype).to_dp()
            field.constraints.required = True
            schema.fields.append(field)
            schema.primaryKey.append(name)

        # Fields
        for name, dtype in self.df.dtypes.items():  # type: ignore
            dvalue: Any = self.df[name].iloc[0] if len(self.df) else None  # type: ignore
            field = PandasField(name=str(name), dtype=dtype, dvalue=dvalue).to_dp()
            schema.fields.append(field)

        return schema

    @classmethod
    def from_dp(cls, schema: Schema) -> PandasSchema:
        columns: Dict[str, pd.Series[Any]] = {}

        # Fields
        for field in schema.fields:
            pandas_field = PandasField.from_dp(field)
            columns[pandas_field.name] = pd.Series(dtype=pandas_field.dtype)  # type: ignore

        # Primary key
        index = schema.primaryKey

        return PandasSchema(df=pd.DataFrame(columns, index=index))



from __future__ import annotations

import os
from functools import lru_cache
from typing import List

from jsonschema.validators import validator_for  # type: ignore

from .. import types
from ..error import Error
from ..errors.metadata import MetadataError
from .data import load_data
from .file import read_file


def select_profile(*, metadata_type: types.IMetadataType) -> str:
    if metadata_type == "package":
        return "data-package"
    elif metadata_type == "resource":
        return "data-resource"
    elif metadata_type == "dialect":
        return "table-dialect"
    elif metadata_type == "schema":
        return "table-schema"
    raise Error(f'Invalid metadata type "{metadata_type}"')


@lru_cache
def read_profile(*, metadata_type: types.IMetadataType) -> types.IDict:
    format = "json"
    name = select_profile(metadata_type=metadata_type)
    path = os.path.join(os.path.dirname(__file__), "..", "profiles", f"{name}.{format}")
    try:
        text = read_file(path)
        data = load_data(text, format=format)
    except Exception:
        raise Error(f'Cannot read profile "{name}" at "{path}"')
    return data


def check_metadata_against_jsonschema(
    metadata: types.IDict, jsonSchema: types.IDict
) -> List[MetadataError]:
    Validator = validator_for(jsonSchema)  # type: ignore
    validator = Validator(jsonSchema)  # type: ignore
    errors: List[MetadataError] = []
    for validation_error in validator.iter_errors(metadata):  # type: ignore
        errors.append(MetadataError(validation_error))  # type: ignore
    return errors



from dplib.models import Package, Resource
from dplib.plugins.zenodo.models import ZenodoPackage


def test_zenodo_package():
    package = ZenodoPackage.from_path("data/plugins/zenodo/package.json")
    assert package.id == "5770714"
    assert len(package.pids) == 2
    assert package.pids["doi"].identifier == "10.5281/zenodo.5770714"
    assert package.created == "2021-12-10T05:47:07.709885+00:00"
    assert package.updated == "2021-12-14T09:04:24.503567+00:00"
    assert package.links["doi"] == "https://doi.org/10.5281/zenodo.5770714"
    assert package.metadata.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.metadata.version == "v1"
    assert len(package.metadata.subjects) == 8
    assert package.metadata.subjects[0].subject == "fishways"
    assert len(package.metadata.creators) == 1
    assert len(package.files.entries) == 2
    resource = package.files.entries["Fishway_Obstruction_Data_v1.csv"]
    assert resource.key == "Fishway_Obstruction_Data_v1.csv"
    assert resource.size == 1377


def test_zenodo_package_to_dp():
    package = ZenodoPackage.from_path("data/plugins/zenodo/package.json").to_dp()
    assert package.id == "https://doi.org/10.5281/zenodo.5770714"
    assert package.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.description
    assert package.description.startswith("<p>This dataset contains pool-weir")
    assert package.homepage == "https://zenodo.org/api/records/5770714"
    assert package.version == "v1"
    assert package.created == "2021-12-10T05:47:07.709885+00:00"
    assert len(package.contributors) == 1
    assert package.contributors[0].title == "Fuentes-Pérez, Juan Francisco"
    assert package.contributors[0].role == "personal"
    assert (
        package.contributors[0].organization
        == "Department of Hydraulics and Hydrology, ETSIIAA, University of Valladolid, 34004 Palencia, Spain"
    )
    assert package.keywords == [
        "fishways",
        "hydraulics",
        "smart fishways",
        "pool-weir",
        "hydrological variability",
        "nonuniformity",
        "clogging",
        "water-level sensors",
    ]
    assert len(package.resources) == 2
    assert package.resources[0].name == "fishway_obstruction_data_v1"
    assert package.resources[0].path == "Fishway_Obstruction_Data_v1.csv"
    assert package.resources[0].format == "csv"
    assert package.resources[0].mediatype == "text/csv"
    assert package.resources[0].bytes == 1377
    assert package.resources[0].hash == "7bdef6756c84c3aea749f8211c557684"
    assert package.resources[1].name == "readme"
    assert package.resources[1].path == "readme.md"
    assert package.resources[1].format == "md"
    assert package.resources[1].mediatype == "application/octet-stream"
    assert package.resources[1].bytes == 1577
    assert package.resources[1].hash == "a23a3c99befca45e706c9343e39f5926"


def test_zenodo_package_from_dp():
    package = ZenodoPackage.from_dp(Package.from_path("data/package-full.json"))
    assert package.metadata.title == "title"
    assert package.metadata.version == "1.0"
    assert package.metadata.description == "description"
    assert len(package.metadata.subjects) == 2
    assert package.metadata.subjects[0].subject == "keyword1"
    assert package.metadata.subjects[1].subject == "keyword2"
    assert package.files.entries["table.csv"].key == "table.csv"


def test_zenodo_package_from_dp_round_trip():
    package = ZenodoPackage.from_dp(
        ZenodoPackage.from_path("data/plugins/zenodo/package.json").to_dp()
    )
    assert package.metadata.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.metadata.version == "v1"
    assert len(package.metadata.subjects) == 8
    assert package.metadata.subjects[0].subject == "fishways"
    assert len(package.files.entries) == 2
    resource = package.files.entries["Fishway_Obstruction_Data_v1.csv"]
    assert resource.key == "Fishway_Obstruction_Data_v1.csv"
    assert resource.size == 1377


def test_zenodo_package_from_dp_inline_resource():
    package = ZenodoPackage.from_dp(Package(resources=[Resource(data=[])]))
    assert package.to_dict() == {}



from __future__ import annotations

from pydantic import BaseModel
from typing_extensions import Self


class Hash(BaseModel):
    type: str
    value: str

    @property
    def short(self):
        return self.value if self.type == "md5" else self.long

    @property
    def long(self):
        return f"{self.type}:{self.value}"

    # Converters

    @classmethod
    def from_text(cls, text: str) -> Self:
        parts = text.split(":", maxsplit=1)
        if len(parts) == 1:
            return cls(type="md5", value=parts[0])
        return cls(type=parts[0], value=parts[1])



# type: ignore
import os

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--ci",
        action="store_true",
        dest="ci",
        default=False,
        help="enable integrational tests",
    )


@pytest.fixture(scope="module")
def vcr_cassette_dir():
    return os.path.join("data", "cassettes")



from __future__ import annotations

import datetime
from typing import Any, Optional

import isodate  # type: ignore
import numpy as np
import pandas as pd
import pandas.core.dtypes.api as pdc

from dplib.error import Error
from dplib.model import Model
from dplib.models import Field


class PandasField(Model, arbitrary_types_allowed=True):
    name: str
    dtype: Any
    dvalue: Optional[Any] = None

    # Converters

    def to_dp(self) -> Field:
        field = Field(name=self.name)

        # Type
        if pdc.is_bool_dtype(self.dtype):  # type: ignore
            field.type = "boolean"
        elif pdc.is_datetime64_any_dtype(self.dtype):  # type: ignore
            field.type = "datetime"
        elif pdc.is_integer_dtype(self.dtype):  # type: ignore
            field.type = "integer"
        elif pdc.is_numeric_dtype(self.dtype):  # type: ignore
            field.type = "number"
        elif self.dvalue is not None:
            if isinstance(self.dvalue, (list, tuple)):  # type: ignore
                field.type = "array"
            elif isinstance(self.dvalue, datetime.datetime):
                field.type = "datetime"
            elif isinstance(self.dvalue, datetime.date):
                field.type = "date"
            elif isinstance(self.dvalue, isodate.Duration):  # type: ignore
                field.type = "duration"
            elif isinstance(self.dvalue, dict):
                field.type = "object"
            elif isinstance(self.dvalue, str):
                field.type = "string"
            elif isinstance(self.dvalue, datetime.time):
                field.type = "time"

        return field

    @classmethod
    def from_dp(cls, field: Field) -> PandasField:
        if not field.name:
            raise Error(f"Field name is required to convert to pandas: {field}")

        # Type
        dtype = np.dtype("O")
        if field.type == "array":
            dtype = np.dtype(list)  # type: ignore
        elif field.type == "boolean":
            dtype = np.dtype(bool)
        elif field.type == "datetime":
            dtype = pd.DatetimeTZDtype(tz="UTC")
        elif field.type == "integer":
            dtype = np.dtype(int)
        elif field.type == "geojson":
            dtype = np.dtype(dict)
        elif field.type == "number":
            dtype = np.dtype(float)
        elif field.type == "object":
            dtype = np.dtype(dict)
        elif field.type == "string":
            dtype = np.dtype(str)
        elif field.type == "year":
            dtype = np.dtype(int)

        return PandasField(name=field.name, dtype=dtype)



from __future__ import annotations

from typing import List, Union

from ... import types
from ...errors.metadata import MetadataError
from ...helpers.data import read_data
from ...helpers.path import infer_basepath
from ...models import Package
from ..metadata.check import check_metadata


def check_package(package: Union[str, types.IDict, Package]) -> List[MetadataError]:
    basepath = None
    if isinstance(package, str):
        basepath = infer_basepath(package)
        package = read_data(package)
    if isinstance(package, Package):
        basepath = package.basepath
        package = package.to_dict()

    # Dereference resources[].dialect/schema
    resources = package.get("resources", [])
    if isinstance(resources, list):
        for resource in resources:  # type: ignore
            for name in ["dialect", "schema"]:
                value = resource.get(name)  # type: ignore
                if value and isinstance(value, str):
                    resource[name] = read_data(value, basepath=basepath)

    return check_metadata(package, type="package")



from __future__ import annotations

import re

from jsonschema.exceptions import ValidationError

from ..error import Error


class MetadataError(Error):
    source: ValidationError
    message: str
    schema_path: str
    object_path: str
    full_message: str

    def __init__(self, source: ValidationError):
        self.source = source
        self.schema_path = "/" + "/".join(map(str, source.schema_path))
        self.object_path = "/" + "/".join(map(str, source.path))
        self.message = re.sub(r"\s+", " ", source.message)
        self.full_message = self.message
        if self.object_path:
            self.full_message = f"[{self.object_path}] {self.message}"
        super().__init__(self.full_message)



from __future__ import annotations

# General

DEFAULT_DIALECT = "postgresql"
ROW_NUMBER_IDENTIFIER = "_rowNumber"
ROW_VALID_IDENTIFIER = "_rowValid"
METADATA_IDENTIFIERS = [ROW_NUMBER_IDENTIFIER, ROW_VALID_IDENTIFIER]



from __future__ import annotations

from typing import Dict, Optional

from dplib.model import Model

from .resource import ZenodoResource


class ZenodoFiles(Model):
    entries: Dict[str, ZenodoResource] = {}
    count: Optional[int] = None
    total_bytes: Optional[int] = None



from typing import Optional

from ...model import Model
from ..metadata.convert import INotation, convert_metadata


def convert_package(
    path: str,
    *,
    format: Optional[str] = None,
    source: Optional[INotation] = None,
    target: Optional[INotation] = None,
) -> Model:
    return convert_metadata(
        path, type="package", format=format, source=source, target=target
    )



from typing import List

from rich.console import Console
from rich.markup import escape
from rich.table import Table

from dplib.errors.metadata import MetadataError


def print_check_results(path: str, errors: List[MetadataError]):
    console = Console()
    if not errors:
        console.print(f"[bold green]Valid: {path}")
        return
    console.print(f"[bold red]Invalid: {path}", style="red")
    console.rule("[bold white]Errors", style="white")
    view = Table()
    view.add_column("location")
    view.add_column("message")
    for error in errors:
        view.add_row(escape(error.object_path), escape(error.message))
    console.print(view)



from __future__ import annotations

import os
import shutil
import tempfile
from pathlib import Path
from typing import Any, Optional

import fsspec  # type: ignore

from ..error import Error
from .path import join_basepath


def read_file(
    path: str,
    *,
    mode: str = "rt",
    encoding: str = "utf-8",
    basepath: Optional[str] = None,
) -> str:
    try:
        path = join_basepath(path, basepath)
        with fsspec.open(path, mode=mode, encoding=encoding) as file:  # type: ignore
            return file.read()  # type: ignore
    except Exception:
        raise Error(f'Cannot read file "{path}"')


def write_file(path: str, body: Any, *, mode: str = "wt", encoding: str = "utf-8"):
    try:
        eff_enc = encoding if mode == "wt" else None
        with tempfile.NamedTemporaryFile(mode, delete=False, encoding=eff_enc) as file:
            file.write(body)
            file.flush()
        move_file(file.name, path, mode=0o644)
    except Exception:
        raise Error(f'Cannot write file "{path}"')


def move_file(source: str, target: str, *, mode: Optional[int] = None):
    try:
        Path(target).parent.mkdir(parents=True, exist_ok=True)
        shutil.move(source, target)
        if mode:
            os.chmod(target, 0o644)
    except Exception:
        raise Error(f'Cannot move file "{source}:{target}"')



from dplib.models import Package
from dplib.plugins.github.models import GithubPackage, GithubResource


def test_github_package():
    package = GithubPackage.from_path("data/plugins/github/package.json")
    assert package.private is False
    assert package.name == "Hello-World"
    assert package.full_name == "octocat/Hello-World"
    assert package.html_url == "https://github.com/octocat/Hello-World"
    assert package.description == "This your first repo!"
    assert package.pushed_at == "2011-01-26T19:06:43Z"
    assert package.created_at == "2011-01-26T19:01:12Z"
    assert package.updated_at == "2011-01-26T19:14:43Z"
    assert package.topics == ["octocat", "atom", "electron", "api"]
    assert package.license
    assert package.license.key == "mit"
    assert package.owner
    assert package.owner.login == "octocat"


def test_github_package_to_dp():
    package = GithubPackage.from_path("data/plugins/github/package.json").to_dp()
    assert package.name == "octocat/Hello-World"
    assert package.title == "Hello-World"
    assert package.description == "This your first repo!"
    assert package.homepage == "https://github.com/octocat/Hello-World"
    assert len(package.licenses) == 1
    assert package.licenses[0].name == "MIT"
    assert package.licenses[0].title == "MIT License"
    assert package.keywords == ["octocat", "atom", "electron", "api"]
    assert package.created == "2011-01-26T19:01:12Z"


def test_github_package_to_dp_with_resources():
    github = GithubPackage.from_path("data/plugins/github/package.json")
    github.resources.append(
        GithubResource(name="table.csv", path="table.csv", size=100, sha="hash")
    )
    package = github.to_dp()
    assert len(package.resources) == 1
    resource = package.resources[0]
    assert resource.name == "table"
    assert resource.path == "table.csv"
    assert resource.bytes == 100
    assert resource.hash == "sha1:hash"


def test_ckan_package_from_dp():
    package = GithubPackage.from_dp(Package.from_path("data/package-full.json"))
    assert package.name == "title"
    assert package.description == "description"
    assert package.topics == ["keyword1", "keyword2"]
    assert package.license
    assert package.license.name == "title"
    assert package.license.spdx_id == "name"
    assert package.license.html_url == "path"
    assert len(package.resources) == 1
    assert package.resources[0].name == "table.csv"
    assert package.resources[0].path == "table.csv"


def test_github_package_from_dp_round_trip():
    package = GithubPackage.from_dp(
        GithubPackage.from_path("data/plugins/github/package.json").to_dp()
    )
    assert package.name == "Hello-World"
    assert package.description == "This your first repo!"
    assert package.topics == ["octocat", "atom", "electron", "api"]
    assert package.license
    assert package.license.name == "MIT License"



from .contributor import Contributor
from .dialect import Dialect
from .license import License
from .package import Package
from .profile import Profile
from .resource import Resource
from .schema import Field, ForeignKey, ForeignKeyReference, Schema
from .source import Source



from __future__ import annotations

from dplib import settings

from .main import program


@program.command(name="version", help="Print the version of the program.")
def command():
    print(settings.VERSION)



from __future__ import annotations

from typing import Optional

from dplib.model import Model


class DataciteRights(Model):
    rights: Optional[str] = None
    rightsUri: Optional[str] = None
    rightsIdentifier: Optional[str] = None
    rightsIdentifierScheme: Optional[str] = None
    lang: Optional[str] = None



from __future__ import annotations

from ..options.system import Debug
from ..program import Program
from . import dialect, package, resource, schema

program = Program()
program.add_typer(dialect.program)
program.add_typer(resource.program)
program.add_typer(package.program)
program.add_typer(schema.program)


@program.callback()
def main(
    debug: bool = Debug,
):
    """
    Python implementation of the Data Package standard and
    various tools for working with data
    """
    program.debug = debug



from __future__ import annotations

from typing import List, Union

from ... import types
from ...errors.metadata import MetadataError
from ...models import Schema
from ..metadata.check import check_metadata


def check_schema(schema: Union[str, types.IDict, Schema]) -> List[MetadataError]:
    if isinstance(schema, Schema):
        schema = schema.to_dict()
    return check_metadata(schema, type="schema")



from typer.testing import CliRunner

from dplib import __version__
from dplib.plugins.cli.commands.main import program

runner = CliRunner()


def test_program_version():
    result = runner.invoke(program, "version")
    assert result.exit_code == 0
    assert result.stdout.count(__version__)



from __future__ import annotations

from typing import List, Union

from ... import types
from ...errors.metadata import MetadataError
from ...helpers.data import read_data
from ...helpers.path import infer_basepath
from ...models import Resource
from ..metadata.check import check_metadata


def check_resource(resource: Union[str, types.IDict, Resource]) -> List[MetadataError]:
    basepath = None
    if isinstance(resource, str):
        basepath = infer_basepath(resource)
        resource = read_data(resource)
    if isinstance(resource, Resource):
        basepath = resource.basepath
        resource = resource.to_dict()

    # Dereference dialect/schema
    for name in ["dialect", "schema"]:
        value = resource.get(name)
        if value and isinstance(value, str):
            resource[name] = read_data(value, basepath=basepath)

    return check_metadata(resource, type="resource")



from rdflib import Namespace
from rdflib.namespace import FOAF, RDF

ADMS = Namespace("http://www.w3.org/ns/adms#")
DCAT = Namespace("http://www.w3.org/ns/dcat#")
DCT = Namespace("http://purl.org/dc/terms/")
OWL = Namespace("http://www.w3.org/2002/07/owl#")

ACCESS_URL = DCAT.accessURL
ACCURAL_PERIODICITY = DCT.accrualPeriodicity
ALTERNATE_IDENTIFIER = ADMS.identifier
BYTE_SIZE = DCAT.byteSize
COMFORMS_TO = DCT.conformsTo
DATASET = DCAT.Dataset
DESCRIPTION = DCT.description
DISTRIBUTION = DCAT.distribution
DOWNLOAD_URL = DCAT.downloadURL
HAS_VERSION = DCT.hasVersion
HOMEPAGE = FOAF.homepage
IDENTIFIER = DCT.identifier
ISSUED = DCT.issued
IS_VERSION_OF = DCT.isVersionOf
KEYWORD = DCAT.keyword
LANDING_PAGE = DCAT.landingPage
LANGUAGE = DCT.language
LICENSE = DCT.license
MEDIA_TYPE = DCAT.mediaType
MODIFIED = DCT.modified
PAGE = FOAF.page
PROVENANCE = DCT.provenance
RELATED_RESOURCE = DCT.relation
SAMPLE = ADMS.sample
SOURCE = DCT.source
THEME = DCAT.theme
TITLE = DCT.title
TYPE = RDF.type
VERSION = OWL.versionInfo

BINDINGS = {
    "adms": ADMS,
    "dcat": DCAT,
    "dct": DCT,
    "owl": OWL,
}



import pytest
from pydantic import ValidationError

from dplib.models import Dialect, Package, Resource, Schema


def test_package_from_path():
    package = Package.from_path("data/package.json")
    assert package.name == "name"
    assert len(package.resources) == 1
    assert package.resources[0].name == "name"
    assert package.resources[0].path == "table.csv"


def test_package_from_path_full():
    package = Package.from_path("data/package-full.json")
    assert package.name == "name"
    assert package.title == "title"
    assert package.description == "description"
    assert package.version == "1.0"
    assert package.homepage == "http://example.com"
    assert package.licenses[0].path == "path"
    assert package.sources[0].path == "path"
    assert package.contributors[0].path == "path"
    assert package.keywords[0] == "keyword1"
    assert package.image == "http://example.com/image.png"
    assert package.created == "2017-01-01T00:00:00Z"


def test_package_from_text():
    text = '{"name": "name"}'
    package = Package.from_text(text, format="json")
    assert package.name == "name"


def test_package_from_dict():
    data = {"name": "name"}
    package = Package.from_dict(data)
    assert package.name == "name"


def test_package_from_dict_invalid():
    data = {"name": 1}
    with pytest.raises(ValidationError):
        Package.from_dict(data)


def test_package_set_proprty_invalid():
    package = Package()
    with pytest.raises(ValidationError):
        package.name = 1  # type: ignore


@pytest.mark.vcr
def test_package_profile():
    package = Package.from_path("data/package-full.json")
    profile = package.get_profile()
    assert profile
    assert profile.jsonSchema.get("title") == "Data Package"


def test_package_get_resource_by_name():
    package = Package.from_path("data/package.json")
    resource = package.get_resource(name="name")
    assert resource
    assert resource.name == "name"
    assert resource.path == "table.csv"


def test_package_get_resource_by_path():
    package = Package.from_path("data/package.json")
    resource = package.get_resource(path="table.csv")
    assert resource
    assert resource.name == "name"
    assert resource.path == "table.csv"


def test_package_resources_basepath():
    package = Package.from_path("data/package.json")
    resource = package.get_resource(name="name")
    assert resource
    assert resource.basepath == "data"


def test_package_add_resource():
    package = Package()
    package.add_resource(Resource(name="name", path="table.csv"))
    resource = package.get_resource(name="name")
    assert resource
    assert resource.name == "name"
    assert resource.path == "table.csv"


def test_package_add_resource_with_basepath():
    package = Package(basepath="data")
    package.add_resource(Resource(name="name", path="table.csv"))
    resource = package.get_resource(name="name")
    assert resource
    assert resource.name == "name"
    assert resource.path == "table.csv"
    assert resource.basepath == "data"


def test_package_dereference():
    package = Package.from_path("data/package-full.json")
    package.dereference()
    resource = package.get_resource(name="name")
    assert resource
    assert isinstance(resource.dialect, Dialect)
    assert resource.dialect.delimiter == ";"
    assert isinstance(resource.schema, Schema)
    assert len(resource.schema.fields) == 2



import pytest
from pydantic import ValidationError

from dplib.models import Dialect


def test_dialect_from_path():
    dialect = Dialect.from_path("data/dialect.json")
    assert dialect.delimiter == ";"
    assert dialect.get_delimiter() == ";"


def test_dialect_from_path_full():
    dialect = Dialect.from_path("data/dialect-full.json")
    assert dialect.delimiter == ";"
    assert dialect.lineTerminator == "\r\n"
    assert dialect.quoteChar == "'"
    assert dialect.doubleQuote is True
    assert dialect.escapeChar == "\\"
    assert dialect.nullSequence == "NULL"
    assert dialect.skipInitialSpace is True
    assert dialect.header is False
    assert dialect.commentChar == "#"


def test_dialect_defaults():
    dialect = Dialect()
    assert dialect.get_delimiter() == ","
    assert dialect.get_line_terminator() == "\r\n"
    assert dialect.get_quote_char() == '"'
    assert dialect.get_double_quote() is True
    assert dialect.get_header() is True


def test_dialect_from_text():
    text = '{"delimiter": ";"}'
    dialect = Dialect.from_text(text, format="json")
    assert dialect.delimiter == ";"


def test_dialect_from_dict():
    data = {"delimiter": ";"}
    dialect = Dialect.from_dict(data)
    assert dialect.delimiter == ";"


def test_dialect_from_dict_invalid():
    data = {"delimiter": 1}
    with pytest.raises(ValidationError):
        Dialect.from_dict(data)


def test_dialect_set_proprty_invalid():
    dialect = Dialect()
    with pytest.raises(ValidationError):
        dialect.delimiter = 1  # type: ignore


@pytest.mark.vcr
def test_dialect_profile():
    dialect = Dialect.from_path("data/dialect-full.json")
    profile = dialect.get_profile()
    assert profile
    assert profile.jsonSchema.get("title") == "CSV Dialect"



from __future__ import annotations

from typing import Optional

from rich.console import Console

from dplib.actions.resource.convert import convert_resource

from ...options.convert import Source, Target
from ...options.path import Format, Path
from .main import program


@program.command(name="convert")
def command(
    path: str = Path,
    format: Optional[str] = Format,
    source: Optional[str] = Source,
    target: Optional[str] = Target,
):
    """Convert a Data Resource descriptor from one notation to another."""
    console = Console()
    model = convert_resource(path, format=format, source=source, target=target)  # type: ignore
    console.print_json(model.to_text(format="json"))



from dplib.models import Package
from dplib.plugins.datacite.models import DatacitePackage


def test_datacite_package():
    package = DatacitePackage.from_path("data/plugins/datacite/package.json")
    assert package.version == "4.2"
    assert package.language == "en-US"
    assert package.publisher == "DataCite"
    assert package.publicationYear == "2014"
    assert package.schemaVersion == "http://datacite.org/schema/kernel-4"
    assert len(package.creators) == 1
    assert package.creators[0].name == "Miller, Elizabeth"
    assert len(package.contributors) == 1
    assert package.contributors[0].name == "Starr, Joan"
    assert len(package.descriptions) == 1
    assert package.descriptions[0].descriptionType == "Abstract"
    assert len(package.identifiers) == 2
    assert package.identifiers[0].identifierType == "DOI"
    assert len(package.rightsList) == 1
    assert package.rightsList[0].lang == "en-US"
    assert len(package.subjects) == 1
    assert package.subjects[0].subject == "000 computer science"
    assert len(package.titles) == 2
    assert package.titles[0].title == "Full DataCite XML Example"


def test_datacite_package_to_dp():
    package = DatacitePackage.from_path("data/plugins/datacite/package.json").to_dp()
    assert package.id == "https://doi.org/https://doi.org/10.1234/example-full"
    assert package.title == "Full DataCite XML Example"
    assert (
        package.description
        == "XML example of all DataCite Metadata Schema v4.3 properties."
    )
    assert (
        package.homepage
        == "https://schema.datacite.org/meta/kernel-4.3/example/datacite-example-full-v4.3.xml"
    )
    assert package.version == "4.2"
    assert package.keywords == ["000 computer science"]
    assert len(package.contributors) == 2
    assert package.contributors[0].title == "Miller, Elizabeth"
    assert package.contributors[1].title == "Starr, Joan"
    assert package.contributors[1].role == "ProjectLeader"
    assert len(package.licenses) == 1
    assert package.licenses[0].path == "http://creativecommons.org/publicdomain/zero/1.0"


def test_datacite_package_from_dp():
    package = DatacitePackage.from_dp(Package.from_path("data/package-full.json"))
    assert package.version == "1.0"
    assert len(package.contributors) == 1
    assert package.contributors[0].name == "title"
    assert len(package.descriptions) == 1
    assert package.descriptions[0].description == "description"
    assert package.descriptions[0].descriptionType == "Abstract"
    assert len(package.identifiers) == 1
    assert package.identifiers[0].identifier == "http://example.com"
    assert package.identifiers[0].identifierType == "URL"
    assert len(package.subjects) == 2
    assert package.subjects[0].subject == "keyword1"
    assert len(package.titles) == 1
    assert package.titles[0].title == "title"


def test_datacite_package_from_dp_round_trip():
    package = DatacitePackage.from_dp(
        DatacitePackage.from_path("data/plugins/datacite/package.json").to_dp()
    )
    assert package.version == "4.2"
    assert len(package.creators) == 1
    assert package.creators[0].name == "Miller, Elizabeth"
    assert len(package.contributors) == 1
    assert package.contributors[0].name == "Starr, Joan"
    assert len(package.descriptions) == 1
    assert package.descriptions[0].descriptionType == "Abstract"
    assert len(package.identifiers) == 2
    assert package.identifiers[0].identifierType == "DOI"
    assert len(package.subjects) == 1
    assert package.subjects[0].subject == "000 computer science"
    assert len(package.titles) == 1
    assert package.titles[0].title == "Full DataCite XML Example"



import pytest
from pydantic import ValidationError

from dplib.models import Field, Schema


def test_schema_from_path():
    schema = Schema.from_path("data/schema.json")
    assert len(schema.fields) == 2
    assert schema.fields[0].name == "id"
    assert schema.fields[0].type == "integer"
    assert schema.fields[1].name == "name"
    assert schema.fields[1].type == "string"


def test_schema_from_path_full():
    schema = Schema.from_path("data/schema-full.json")
    assert len(schema.fields) == 15
    assert len(schema.primaryKey) == 4
    assert len(schema.foreignKeys) == 1
    assert schema.foreignKeys[0].fields == ["position_title"]
    assert schema.foreignKeys[0].reference.resource == "positions"
    assert schema.foreignKeys[0].reference.fields == ["name"]


def test_schema_defaults():
    schema = Schema()
    assert schema.missingValues == [""]


def test_schema_from_text():
    text = '{"missingValues": ["x"]}'
    schema = Schema.from_text(text, format="json")
    assert schema.missingValues == ["x"]


def test_schema_from_dict():
    data = {"missingValues": ["x"]}
    schema = Schema.from_dict(data)
    assert schema.missingValues == ["x"]


def test_schema_from_dict_invalid():
    data = {"missingValues": 1}
    with pytest.raises(ValidationError):
        Schema.from_dict(data)


def test_schema_set_proprty_invalid():
    schema = Schema()
    with pytest.raises(ValidationError):
        schema.missingValues = 1  # type: ignore


@pytest.mark.vcr
def test_schema_profile():
    schema = Schema.from_path("data/schema-full.json")
    profile = schema.get_profile()
    assert profile
    assert profile.jsonSchema.get("title") == "Table Schema"


def test_schema_add_field():
    schema = Schema()
    schema.add_field(Field(name="id", type="integer"))
    field = schema.get_field(name="id")
    assert field
    assert field.name == "id"
    assert field.type == "integer"


def test_schema_get_field():
    schema = Schema.from_path("data/schema.json")
    field = schema.get_field(name="id")
    assert field
    assert field.name == "id"
    assert field.type == "integer"


def test_schema_to_dict():
    schema = Schema()
    assert schema.to_dict() == {}
    schema.missingValues.append("x")
    assert schema.to_dict() == {"missingValues": ["", "x"]}



from __future__ import annotations

from typing import Literal, Union

FieldType = Union[
    Literal["any"],
    Literal["array"],
    Literal["boolean"],
    Literal["date"],
    Literal["datetime"],
    Literal["duration"],
    Literal["geojson"],
    Literal["geopoint"],
    Literal["integer"],
    Literal["number"],
    Literal["object"],
    Literal["string"],
    Literal["time"],
    Literal["year"],
    Literal["yearmonth"],
]



from dplib.actions.dialect.check import check_dialect
from dplib.models import Dialect


def test_check_dialect():
    errors = check_dialect("data/dialect.json")
    assert len(errors) == 0


def test_check_dialect_invalid():
    errors = check_dialect("data/dialect-invalid.json")
    assert len(errors) == 1
    error = errors[0]
    assert error.message == "1 is not of type 'string'"
    assert error.schema_path == "/properties/delimiter/type"
    assert error.object_path == "/delimiter"
    assert error.full_message == "[/delimiter] 1 is not of type 'string'"


def test_check_dialect_from_model():
    dialect = Dialect(delimiter=";")
    errors = check_dialect(dialect)
    assert len(errors) == 0



from __future__ import annotations

from typing import Optional

from dplib.model import Model


class CkanTag(Model):
    name: str
    id: Optional[str] = None
    display_name: Optional[str] = None



import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg

from dplib.models import Field, Schema
from dplib.plugins.sql.models import SqlSchema


def test_sql_schema_from_dp():
    schema = SqlSchema.from_dp(
        Schema.from_path("data/schema-types.json"), table_name="test"
    )
    assert schema.table.name == "test"
    assert len(schema.table.columns) == 16
    assert schema.get_field_names() == [
        "any",
        "array",
        "boolean",
        "date",
        "date_year",
        "datetime",
        "duration",
        "geojson",
        "geopoint",
        "integer",
        "number",
        "object",
        "string",
        "time",
        "year",
        "yearmonth",
    ]
    assert schema.get_field_types() == [
        sa.Text,
        pg.JSONB,
        sa.Boolean,
        sa.Date,
        sa.Date,
        sa.DateTime,
        sa.Text,
        pg.JSONB,
        sa.Text,
        sa.Integer,
        sa.Numeric,
        pg.JSONB,
        sa.Text,
        sa.Time,
        sa.Integer,
        sa.Text,
    ]


def test_sql_schema_to_dp_round_trip():
    schema = SqlSchema.from_dp(
        Schema.from_path("data/schema-types.json"), table_name="test"
    ).to_dp()
    assert len(schema.fields) == 16
    assert schema.get_field_names() == [
        "any",
        "array",
        "boolean",
        "date",
        "date_year",
        "datetime",
        "duration",
        "geojson",
        "geopoint",
        "integer",
        "number",
        "object",
        "string",
        "time",
        "year",
        "yearmonth",
    ]
    assert schema.get_field_types() == [
        "string",
        "object",
        "boolean",
        "date",
        "date",
        "datetime",
        "string",
        "object",
        "string",
        "integer",
        "number",
        "object",
        "string",
        "time",
        "integer",
        "string",
    ]


def test_sql_schema_from_dp_with_string_constraints():
    field = Field(name="string", type="string")
    field.constraints.minLength = 1
    field.constraints.enum = ["a", "b", "c"]
    field.constraints.pattern = "^[a-z]+$"
    schema = SqlSchema.from_dp(Schema(fields=[field]), table_name="test")
    assert len(schema.table.columns[0].constraints) == 2
    assert isinstance(schema.table.columns[0].type, sa.Enum)



from __future__ import annotations

from typing import List, Optional

import pydantic

from dplib.model import Model


class ZenodoCreatorAffiliation(Model):
    name: str


class ZenodoCreatorPersonOrOrg(Model):
    family_name: Optional[str] = None
    given_name: Optional[str] = None
    name: Optional[str] = None
    type: Optional[str] = None


class ZenodoCreator(Model):
    affiliations: List[ZenodoCreatorAffiliation] = []
    person_or_org: ZenodoCreatorPersonOrOrg = pydantic.Field(
        default_factory=ZenodoCreatorPersonOrOrg
    )



from __future__ import annotations

from typing import Optional

from dplib.model import Model


class DataciteDescription(Model):
    description: str
    descriptionType: str
    lang: Optional[str] = None



from __future__ import annotations

from typing import Optional

from dplib.model import Model


class DataciteTitle(Model):
    title: str
    titleType: Optional[str] = None
    lang: Optional[str] = None



from __future__ import annotations

from typing import Any, List, Optional, Union

import pydantic

from ... import types
from ...helpers.file import join_basepath
from ...model import Model
from ..contributor import Contributor
from ..dialect import Dialect
from ..license import License
from ..profile import Profile
from ..schema import Schema
from ..source import Source
from .hash import Hash


class Resource(Model):
    basepath: Optional[str] = pydantic.Field(default=None, exclude=True)

    name: Optional[str] = None
    type: Optional[str] = None
    path: Optional[Union[str, List[str]]] = None
    data: Optional[Any] = None
    profile: Optional[str] = None
    dialect: Optional[Union[Dialect, str]] = None
    schema: Optional[Union[Schema, str]] = None  # type: ignore

    title: Optional[str] = None
    description: Optional[str] = None
    format: Optional[str] = None
    mediatype: Optional[str] = None
    encoding: Optional[str] = None
    bytes: Optional[int] = None
    hash: Optional[str] = None
    sources: List[Source] = []
    licenses: List[License] = []
    contributors: List[Contributor] = []

    # Getters

    def get_fullpath(self) -> Optional[str]:
        if self.path and isinstance(self.path, str):
            return join_basepath(self.path, self.basepath)

    def get_source(self) -> Optional[Union[str, types.IDict]]:
        return self.data if self.data is not None else self.get_fullpath()

    def get_profile(self) -> Optional[Profile]:
        if self.profile:
            return Profile.from_path(self.profile)

    def get_dialect(self) -> Optional[Dialect]:
        if self.dialect:
            if isinstance(self.dialect, str):
                return Dialect.from_path(self.dialect, basepath=self.basepath)
            return self.dialect

    def get_schema(self) -> Optional[Schema]:
        if self.schema:
            if isinstance(self.schema, str):
                return Schema.from_path(self.schema, basepath=self.basepath)
            return self.schema

    def get_hash(self) -> Optional[Hash]:
        if self.hash:
            return Hash.from_text(self.hash)

    # Methods

    def dereference(self):
        if isinstance(self.dialect, str):
            self.dialect = Dialect.from_path(self.dialect, basepath=self.basepath)
        if isinstance(self.schema, str):
            self.schema = Schema.from_path(self.schema, basepath=self.basepath)  # type: ignore



from __future__ import annotations

from typing import List, Optional

from ...model import Model
from ..profile import Profile
from .field import Field
from .foreignKey import ForeignKey


class Schema(Model):
    """Schema model"""

    profile: Optional[str] = None

    fields: List[Field] = []
    """List of fields"""

    missingValues: List[str] = [""]
    primaryKey: List[str] = []
    foreignKeys: List[ForeignKey] = []

    # Getters

    def get_profile(self) -> Optional[Profile]:
        if self.profile:
            return Profile.from_path(self.profile)

    def get_field(self, *, name: Optional[str] = None) -> Optional[Field]:
        for field in self.fields:
            if name and field.name == name:
                return field

    def get_field_names(self) -> List[str]:
        names: List[str] = []
        for field in self.fields:
            names.append(field.name or "")
        return names

    def get_field_types(self) -> List[str]:
        types: List[str] = []
        for field in self.fields:
            types.append(field.type)
        return types

    # Setters

    def add_field(self, field: Field) -> None:
        self.fields.append(field)



from __future__ import annotations

from typing import Any, List, Optional

from ...model import Model


class Constraints(Model):
    required: Optional[bool] = None
    unique: Optional[bool] = None
    enum: Optional[List[Any]] = None
    minimum: Optional[Any] = None
    maximum: Optional[Any] = None
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    pattern: Optional[str] = None



import pytest

from dplib.actions.package.check import check_package
from dplib.models import Package


def test_check_package():
    errors = check_package("data/package.json")
    assert len(errors) == 0


def test_check_package_invalid():
    errors = check_package("data/package-invalid.json")
    assert len(errors) == 1
    error = errors[0]
    assert error.message == "1 is not of type 'string'"
    assert error.schema_path == "/properties/name/type"
    assert error.object_path == "/name"
    assert error.full_message == "[/name] 1 is not of type 'string'"


def test_check_package_invalid_dereferencing():
    errors = check_package("data/package-invalid-dereferencing.json")
    assert len(errors) == 1
    error = errors[0]
    assert (
        error.full_message == "[/resources/0/dialect/delimiter] 1 is not of type 'string'"
    )


@pytest.mark.vcr
def test_check_package_custom_profile():
    errors = check_package("data/package-custom-profile.json")
    assert len(errors) == 1
    error = errors[0]
    assert error.full_message == "[/] 'requiredProperty' is a required property"


def test_check_package_from_model():
    package = Package(name="name")
    errors = check_package(package)
    assert len(errors) == 1



from __future__ import annotations

from typing import List, Optional

from dplib.model import Model
from dplib.models import Contributor, License, Package

from .contributor import DataciteContributor, DataciteContributorAffiliation
from .description import DataciteDescription
from .identifier import DataciteIdentifier
from .rights import DataciteRights
from .subject import DataciteSubject
from .title import DataciteTitle

# References:
# - https://github.com/inveniosoftware/datacite/blob/master/tests/data/datacite-v4.3-full-example.json
# - https://zenodo.org/records/5770714 (export Datacite JSON)
# - https://schema.datacite.org/meta/kernel-4.3/doc/DataCite-MetadataKernel_v4.3.pdf


class DatacitePackage(Model):
    version: Optional[str] = None
    language: Optional[str] = None
    publisher: Optional[str] = None
    publicationYear: Optional[str] = None
    schemaVersion: Optional[str] = None

    creators: List[DataciteContributor] = []
    contributors: List[DataciteContributor] = []
    descriptions: List[DataciteDescription] = []
    identifiers: List[DataciteIdentifier] = []
    rightsList: List[DataciteRights] = []
    subjects: List[DataciteSubject] = []
    titles: List[DataciteTitle] = []

    # Converters

    def to_dp(self) -> Package:
        package = Package()

        # Id
        for identifier in self.identifiers:
            if identifier.identifierType == "DOI":
                package.id = f"https://doi.org/{identifier.identifier}"
                break

        # Version
        if self.version:
            package.version = self.version

        # Title
        for title in self.titles:
            if not title.titleType:
                package.title = title.title
                break

        # Description
        for description in self.descriptions:
            if description.descriptionType == "Abstract":
                package.description = description.description
                break

        # Homepage
        for identifier in self.identifiers:
            if identifier.identifierType == "URL":
                package.homepage = identifier.identifier
                break

        # Keywords
        for subject in self.subjects:
            package.keywords.append(subject.subject)

        # Contributors
        for type, items in [("creator", self.creators), ("other", self.contributors)]:
            for item in items:
                contributor = Contributor(title=item.name)
                contributor.role = type if type == "creator" else item.contributorType
                for affiliation in item.affiliation:
                    contributor.organization = affiliation.name
                    break
                package.contributors.append(contributor)

        # Licenses
        for rights in self.rightsList:
            if rights.rightsIdentifier or rights.rightsUri:
                license = License(name=rights.rightsIdentifier, path=rights.rightsUri)
                if rights.rights:
                    license.title = rights.rights
                package.licenses.append(license)

        return package

    @classmethod
    def from_dp(cls, package: Package) -> DatacitePackage:
        datacite = DatacitePackage()

        # Id
        if package.id:
            if package.id.startswith("https://doi.org/"):
                doi = package.id.replace("https://doi.org/", "")
                datacite.identifiers.append(
                    DataciteIdentifier(identifierType="DOI", identifier=doi)
                )

        # Version
        if package.version:
            datacite.version = package.version

        # Title
        if package.title:
            datacite.titles.append(DataciteTitle(title=package.title))

        # Description
        if package.description:
            datacite.descriptions.append(
                DataciteDescription(
                    descriptionType="Abstract",
                    description=package.description,
                )
            )

        # Homepage
        if package.homepage:
            datacite.identifiers.append(
                DataciteIdentifier(identifierType="URL", identifier=package.homepage)
            )

        # Keywords
        for keyword in package.keywords:
            datacite.subjects.append(DataciteSubject(subject=keyword))

        # Contributors
        for contributor in package.contributors:
            item = DataciteContributor(name=contributor.title)
            if contributor.organization:
                org = DataciteContributorAffiliation(name=contributor.organization)
                item.affiliation.append(org)
            if contributor.role:
                item.contributorType = contributor.role
            target = datacite.contributors
            if contributor.role in ["author", "creator"]:
                target = datacite.creators
            target.append(item)

        return datacite



from dplib.actions.resource.check import check_resource
from dplib.models import Resource


def test_check_resource():
    errors = check_resource("data/resource.json")
    assert len(errors) == 0


def test_check_resource_invalid():
    errors = check_resource("data/resource-invalid.json")
    assert len(errors) == 1
    error = errors[0]
    assert error.message == "1 is not of type 'string'"
    assert error.schema_path == "/properties/name/type"
    assert error.object_path == "/name"
    assert error.full_message == "[/name] 1 is not of type 'string'"


def test_check_resource_invalid_dereferencing():
    errors = check_resource("data/resource-invalid-dereferencing.json")
    assert len(errors) == 1
    error = errors[0]
    assert error.full_message == "[/dialect/delimiter] 1 is not of type 'string'"


def test_check_resource_from_model():
    resource = Resource(name="name", path="path")
    errors = check_resource(resource)
    assert len(errors) == 0



from __future__ import annotations

import sys
from typing import Any, Optional

import typer
from rich.console import Console
from rich.markup import escape


class Program(typer.Typer):
    debug: Optional[bool] = False

    def __call__(self, *args: Any, **kwargs: Any):
        try:
            super().__call__(*args, **kwargs)
        except Exception as exception:
            try:
                if self.debug:
                    raise
                console = Console()
                console.print(escape(str(exception)), style="bold red")
                raise typer.Exit(code=1)
            except typer.Exit as e:
                sys.exit(e.exit_code)
            except KeyError:
                raise



from dplib.models import Package, Resource
from dplib.plugins.ckan.models import CkanPackage


def test_ckan_package():
    package = CkanPackage.from_path("data/plugins/ckan/package.json")
    assert package.name == "sample-dataset-1"
    assert package.title == "Sample Dataset"
    assert package.organization
    assert package.organization.name == "sample-organization"
    assert package.version == "1.0"
    assert package.license_id == "cc-by"
    assert len(package.resources) == 9
    assert len(package.tags) == 8
    assert package.resources[0].name == "sample-linked.csv"
    assert package.tags[0].name == "csv"


def test_ckan_package_to_dp():
    package = CkanPackage.from_path("data/plugins/ckan/package.json").to_dp()
    assert package.name == "sample-dataset-1"
    assert package.title == "Sample Dataset"
    assert package.version == "1.0"
    assert package.licenses[0].name == "cc-by"
    assert len(package.resources) == 9
    assert len(package.keywords) == 8
    assert len(package.contributors) == 2
    assert package.resources[0].path == "sample-linked.csv"
    assert package.keywords[0] == "csv"
    assert package.contributors[0].title == "Test Author"
    assert package.contributors[0].role == "author"
    assert package.custom["ckan:id"] == "c322307a-b871-44fe-a602-32ee8437ff04"


def test_ckan_package_from_dp():
    package = CkanPackage.from_dp(Package.from_path("data/package-full.json"))
    assert package.name == "name"
    assert package.name == "name"
    assert package.title == "title"
    assert package.notes == "description"
    assert package.version == "1.0"
    assert package.license_id == "name"
    assert package.license_title == "title"
    assert package.license_url == "path"
    assert len(package.resources) == 1
    assert package.resources[0].name == "table.csv"
    assert len(package.tags) == 2
    assert package.tags[0].name == "keyword1"
    assert package.tags[1].name == "keyword2"


def test_ckan_package_from_dp_round_trip():
    package = CkanPackage.from_dp(
        CkanPackage.from_path("data/plugins/ckan/package.json").to_dp()
    )
    assert package.name == "sample-dataset-1"
    assert package.title == "Sample Dataset"
    assert package.version == "1.0"
    assert package.license_id == "cc-by"
    assert len(package.resources) == 9
    assert len(package.tags) == 8
    assert package.resources[0].name == "sample-linked.csv"
    assert package.tags[0].name == "csv"


def test_ckan_package_from_dp_inline_resource():
    package = CkanPackage.from_dp(Package(resources=[Resource(data=[])]))
    assert package.to_dict() == {}



from __future__ import annotations

import json
from importlib import import_module
from typing import Optional

from .. import types
from ..error import Error
from .file import read_file, write_file
from .path import infer_format


def read_data(
    path: str, *, format: Optional[str] = None, basepath: Optional[str] = None
) -> types.IDict:
    if not format:
        format = infer_format(path, raise_missing=True)
    text = read_file(path, basepath=basepath)
    data = load_data(text, format=format)
    return data


def write_data(path: str, data: types.IDict, *, format: Optional[str] = None):
    if not format:
        format = infer_format(path, raise_missing=True)
    text = dump_data(data, format=format)
    write_file(path, text)


def load_data(text: str, *, format: str) -> types.IDict:
    try:
        if format == "json":
            return json.loads(text)
        elif format == "yaml":
            yaml = import_module("yaml")
            return yaml.safe_load(text)
    except Exception:
        raise Error(f'Cannot load "{format}" data from text: {text}')
    raise Error(f"Cannot load data from text with format: {format}")


def dump_data(data: types.IDict, *, format: str) -> str:
    try:
        if format == "json":
            return json.dumps(data)
        elif format == "yaml":
            yaml = import_module("yaml")
            return yaml.dump(data)
    except Exception:
        raise Error(f'Cannot dump "{format}" text from data: {data}')
    raise Error(f"Cannot dump data to text with format: {format}")


def clean_data(data: types.IDict):
    for key, value in list(data.items()):
        if isinstance(value, dict):
            clean_data(value)  # type: ignore
        elif isinstance(value, list):
            for item in value:  # type: ignore
                if isinstance(item, dict):
                    clean_data(item)  # type: ignore
        if value is None or value == [] or value == {}:
            data.pop(key)



from __future__ import annotations

from typing import Optional

from ...model import Model
from ..profile import Profile


class Dialect(Model):
    profile: Optional[str] = None

    """A dialect description for parsing CSV files"""

    delimiter: Optional[str] = None
    """
    Specifies the character sequence which should separate fields (aka columns).
    """

    lineTerminator: Optional[str] = None
    """
    Specifies the character sequence which should terminate rows.
    """

    quoteChar: Optional[str] = None
    """
    Specifies a one-character string to use as the quoting character.
    """

    doubleQuote: Optional[bool] = None
    """
    Controls the handling of quotes inside fields.
    """

    escapeChar: Optional[str] = None
    """
    Specifies a one-character string to use as the escape character.
    """

    nullSequence: Optional[str] = None
    """
    Specifies the character sequence which represents a null value.
    """

    skipInitialSpace: Optional[bool] = None
    """
    Specifies whether or not parsing will skip initial spaces after the delimiter.
    """

    header: Optional[bool] = None
    """
    Specifies whether or not the file includes a header row.
    """

    commentChar: Optional[str] = None
    """
    Specifies a one-character string to use as the comment character.
    """

    # Getters

    def get_profile(self) -> Optional[Profile]:
        if self.profile:
            return Profile.from_path(self.profile)

    def get_delimiter(self) -> str:
        return self.delimiter if self.delimiter is not None else ","

    def get_line_terminator(self) -> str:
        return self.lineTerminator if self.lineTerminator is not None else "\r\n"

    def get_quote_char(self) -> str:
        return self.quoteChar if self.quoteChar is not None else '"'

    def get_double_quote(self) -> bool:
        return self.doubleQuote if self.doubleQuote is not None else True

    def get_header(self) -> bool:
        return self.header if self.header is not None else True



from typing import Any, List, Optional

from ... import types
from ...model import Model
from .rule import Rule


class Profile(Model):
    jsonSchema: types.IDict = {}

    name: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    version: Optional[str] = None
    rules: List[Rule] = []

    # Converters

    def to_dict(self):
        data: types.IDict = {}
        profile = super().to_dict()
        schema = profile.pop("jsonSchema", {})
        data.update(schema)
        data.update(metadataProfile=profile)
        return data

    @classmethod
    def from_dict(cls, data: types.IDict, **kwargs: Any):
        profile = data.pop("metadataProfile", {})
        return cls(jsonSchema=data, **profile)



from dplib.actions.schema.check import check_schema
from dplib.models import Schema


def test_check_schema():
    errors = check_schema("data/schema.json")
    assert len(errors) == 0


def test_check_schema_invalid():
    errors = check_schema("data/schema-invalid.json")
    assert len(errors) == 2
    error = errors[0]
    assert error.message == "'fields' is a required property"
    assert error.schema_path == "/required"
    assert error.object_path == "/"
    assert error.full_message == "[/] 'fields' is a required property"


def test_check_schema_from_model():
    schema = Schema()
    errors = check_schema(schema)
    assert len(errors) == 1



import polars as pl

from dplib.models import Schema
from dplib.plugins.polars.models import PolarsSchema


def test_polars_schema_from_dp():
    schema = PolarsSchema.from_dp(Schema.from_path("data/schema-types.json"))
    assert len(schema.df.schema) == 16
    assert schema.get_field_names() == [
        "any",
        "array",
        "boolean",
        "date",
        "date_year",
        "datetime",
        "duration",
        "geojson",
        "geopoint",
        "integer",
        "number",
        "object",
        "string",
        "time",
        "year",
        "yearmonth",
    ]
    assert schema.get_field_types() == [
        pl.Utf8,
        pl.List(pl.Null),
        pl.Boolean,
        pl.Date,
        pl.Date,
        pl.Datetime(time_unit="us"),
        pl.Duration(time_unit="us"),
        pl.Struct([pl.Field("", pl.Null)]),
        pl.List(pl.Null),
        pl.Int64,
        pl.Decimal(scale=0),
        pl.Struct([pl.Field("", pl.Null)]),
        pl.Utf8,
        pl.Time,
        pl.Int8,
        pl.List(pl.Null),
    ]


def test_polars_schema_to_dp_round_trip():
    schema = PolarsSchema.from_dp(Schema.from_path("data/schema-types.json")).to_dp()
    assert len(schema.fields) == 16
    assert schema.get_field_names() == [
        "any",
        "array",
        "boolean",
        "date",
        "date_year",
        "datetime",
        "duration",
        "geojson",
        "geopoint",
        "integer",
        "number",
        "object",
        "string",
        "time",
        "year",
        "yearmonth",
    ]
    assert schema.get_field_types() == [
        "string",
        "array",
        "boolean",
        "date",
        "date",
        "datetime",
        "duration",
        "object",
        "array",
        "integer",
        "number",
        "object",
        "string",
        "time",
        "integer",
        "array",
    ]



from dplib.models import Package
from dplib.plugins.dcat.models import DcatPackage


def test_dcat_package():
    package = DcatPackage.from_path("data/plugins/dcat/package.xml")
    assert package.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.version == "v1"
    assert package.description
    assert package.description.startswith("This dataset contains pool-weir type fishway")
    assert package.issued == "2021"
    assert package.is_version_of == ["https://doi.org/10.5281/zenodo.5770713"]
    assert package.identifier == "https://doi.org/10.5281/zenodo.5770714"
    assert package.languages == [
        "http://publications.europa.eu/resource/authority/language/ENG"
    ]
    assert package.pages == ["https://doi.org/10.5281/zenodo.5770714"]
    assert len(package.keywords) == 8
    assert package.keywords[0] == "fishways"
    assert len(package.distributions) == 3
    assert (
        package.distributions[1].download_url
        == "https://zenodo.org/records/5770714/files/Fishway_Obstruction_Data_v1.csv"
    )
    assert package.distributions[1].byte_size == 1377
    assert package.distributions[1].media_type == "text/csv"


def test_dcat_package_to_dp():
    package = DcatPackage.from_path("data/plugins/dcat/package.xml").to_dp()
    assert package.id == "https://doi.org/10.5281/zenodo.5770714"
    assert package.resources[1].bytes == 1577
    assert package.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.description
    assert package.description.startswith("This dataset contains pool-weir type fishway")
    assert package.version == "v1"
    assert len(package.keywords) == 8
    assert package.keywords[0] == "fishways"
    assert len(package.resources) == 2
    assert package.resources[0].name == "fishway_obstruction_data_v1"
    assert (
        package.resources[0].path
        == "https://zenodo.org/records/5770714/files/Fishway_Obstruction_Data_v1.csv"
    )
    assert package.resources[0].mediatype == "text/csv"
    assert package.resources[0].bytes == 1377
    assert package.resources[1].name == "readme"
    assert (
        package.resources[1].path == "https://zenodo.org/records/5770714/files/readme.md"
    )


def test_dcat_package_from_dp():
    package = DcatPackage.from_dp(Package.from_path("data/package-full.json"))
    assert package.identifier == "id"
    assert len(package.distributions) == 1
    assert package.distributions[0].download_url == "table.csv"
    assert package.description == "description"
    assert package.homepage == "http://example.com"
    assert package.issued == "2017-01-01T00:00:00Z"
    assert package.keywords == ["keyword1", "keyword2"]
    assert package.title == "title"
    assert package.version == "1.0"


def test_dcat_package_from_dp_round_trip():
    package = DcatPackage.from_dp(
        DcatPackage.from_path("data/plugins/dcat/package.xml").to_dp()
    )
    assert package.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.version == "v1"
    assert package.description
    assert package.description.startswith("This dataset contains pool-weir type fishway")
    assert package.identifier == "https://doi.org/10.5281/zenodo.5770714"
    assert len(package.keywords) == 8
    assert package.keywords[0] == "fishways"
    assert len(package.distributions) == 2
    assert (
        package.distributions[0].download_url
        == "https://zenodo.org/records/5770714/files/Fishway_Obstruction_Data_v1.csv"
    )
    assert package.distributions[0].byte_size == 1377
    assert package.distributions[0].media_type == "text/csv"


def test_dcat_package_from_graph_round_trip():
    package = DcatPackage.from_graph(
        DcatPackage.from_path("data/plugins/dcat/package.xml").to_graph()
    )
    assert package.title == "Fishway_Obstruction_Data_v1.csv"
    assert package.version == "v1"
    assert package.description
    assert package.description.startswith("This dataset contains pool-weir type fishway")
    assert package.issued == "2021"
    assert package.is_version_of == ["https://doi.org/10.5281/zenodo.5770713"]
    assert package.identifier == "https://doi.org/10.5281/zenodo.5770714"
    assert package.languages == [
        "http://publications.europa.eu/resource/authority/language/ENG"
    ]
    assert package.pages == ["https://doi.org/10.5281/zenodo.5770714"]
    assert len(package.keywords) == 8
    assert package.keywords[0] == "fishways"
    assert len(package.distributions) == 3
    assert (
        package.distributions[1].download_url
        == "https://zenodo.org/records/5770714/files/Fishway_Obstruction_Data_v1.csv"
    )
    assert package.distributions[1].byte_size == 1377
    assert package.distributions[1].media_type == "text/csv"


def test_dcat_package_to_text():
    text = DcatPackage.from_path("data/plugins/dcat/package.xml").to_text(format="xml")
    assert text.startswith("<?xml")



from __future__ import annotations

from typing import Dict, Optional

import pydantic

from dplib.model import Model
from dplib.models import Contributor, Package

from .creator import ZenodoCreator, ZenodoCreatorAffiliation
from .files import ZenodoFiles
from .metadata import ZenodoMetadata
from .pid import ZenodoPid
from .resource import ZenodoResource
from .subject import ZenodoSubject

# References:
# - https://zenodo.org/records/6533675


class ZenodoPackage(Model):
    files: ZenodoFiles = pydantic.Field(default_factory=ZenodoFiles)
    metadata: ZenodoMetadata = pydantic.Field(default_factory=ZenodoMetadata)

    id: Optional[str] = None
    pids: Dict[str, ZenodoPid] = {}
    created: Optional[str] = None
    updated: Optional[str] = None
    links: Dict[str, str] = {}

    # Converters

    def to_dp(self):
        package = Package()

        # Id
        if self.links.get("doi"):
            package.id = self.links.get("doi")

        # Title
        if self.metadata.title:
            package.title = self.metadata.title

        # Description
        if self.metadata.description:
            package.description = self.metadata.description

        # Version
        if self.metadata.version:
            package.version = self.metadata.version

        # Created
        if self.created:
            package.created = self.created

        # Homepage
        if self.links.get("self"):
            package.homepage = self.links.get("self")

        # Resources
        for entry in self.files.entries.values():
            resource = entry.to_dp()
            package.resources.append(resource)

        # Keywords
        for subject in self.metadata.subjects:
            package.keywords.append(subject.subject)

        # Contributors
        for creator in self.metadata.creators:
            if creator.person_or_org.name:
                contributor = Contributor(title=creator.person_or_org.name)
                if creator.person_or_org.type:
                    contributor.role = creator.person_or_org.type
                if creator.affiliations:
                    contributor.organization = creator.affiliations[0].name
                package.contributors.append(contributor)

        return package

    @classmethod
    def from_dp(cls, package: Package) -> ZenodoPackage:
        zenodo = ZenodoPackage()

        # Title
        if package.title:
            zenodo.metadata.title = package.title

        # Description
        if package.description:
            zenodo.metadata.description = package.description

        # Version
        if package.version:
            zenodo.metadata.version = package.version

        # Resources
        for resource in package.resources:
            entry = ZenodoResource.from_dp(resource)
            if entry:
                zenodo.files.entries[entry.key] = entry

        # Keywords
        for keyword in package.keywords:
            subject = ZenodoSubject(subject=keyword)
            zenodo.metadata.subjects.append(subject)

        # Contributors
        for contributor in package.contributors:
            creator = ZenodoCreator()
            creator.person_or_org.name = contributor.title
            if contributor.role:
                creator.person_or_org.type = contributor.role
            if contributor.organization:
                affiliation = ZenodoCreatorAffiliation(name=contributor.organization)
                creator.affiliations.append(affiliation)

        return zenodo



from __future__ import annotations

from typing import Any

import polars as pl

from dplib.error import Error
from dplib.model import Model
from dplib.models import Field


class PolarsField(Model, arbitrary_types_allowed=True):
    name: str
    dtype: Any
    #  dtype: pl.PolarsDataType

    # Converters

    def to_dp(self) -> Field:
        field = Field(name=self.name)

        if self.dtype in ARRAY_TYPES:
            field.type = "array"
        elif self.dtype in BOOLEAN_TYPES:
            field.type = "boolean"
        elif self.dtype in DATE_TYPES:
            field.type = "date"
        elif self.dtype in DATETIME_TYPES:
            field.type = "datetime"
        elif self.dtype in DURATION_TYPES:
            field.type = "duration"
        elif self.dtype in INTEGER_TYPES:
            field.type = "integer"
        elif self.dtype in NUMBER_TYPES:
            field.type = "number"
        elif self.dtype in OBJECT_TYPES:
            field.type = "object"
        elif self.dtype in STRING_TYPES:
            field.type = "string"
        elif self.dtype in TIME_TYPES:
            field.type = "time"

        return field

    @classmethod
    def from_dp(cls, field: Field) -> PolarsField:
        if not field.name:
            raise Error(f"Field name is required to convert to polars: {field}")

        # Type
        dtype = pl.Utf8
        if field.type == "array":
            dtype = pl.List
        elif field.type == "boolean":
            dtype = pl.Boolean
        elif field.type == "date":
            dtype = pl.Date
        elif field.type == "datetime":
            dtype = pl.Datetime
        elif field.type == "duration":
            dtype = pl.Duration
        elif field.type == "geojson":
            dtype = pl.Struct
        elif field.type == "geopoint":
            dtype = pl.List
        elif field.type == "integer":
            dtype = pl.Int64
        elif field.type == "number":
            dtype = pl.Decimal
        elif field.type == "object":
            dtype = pl.Struct
        elif field.type == "string":
            dtype = pl.Utf8
        elif field.type == "time":
            dtype = pl.Time
        elif field.type == "year":
            dtype = pl.Int8
        elif field.type == "yearmonth":
            dtype = pl.List

        return PolarsField(name=field.name, dtype=dtype)


ARRAY_TYPES = (pl.Array, pl.List)
BOOLEAN_TYPES = (pl.Boolean,)
DATE_TYPES = (pl.Date,)
DATETIME_TYPES = (pl.Datetime,)
DURATION_TYPES = (pl.Duration,)
INTEGER_TYPES = (
    pl.Int8,
    pl.Int16,
    pl.Int32,
    pl.Int64,
    pl.UInt8,
    pl.UInt16,
    pl.UInt32,
    pl.UInt64,
)
NUMBER_TYPES = (pl.Float32, pl.Float64, pl.Decimal)
STRING_TYPES = (pl.Utf8, pl.Categorical)
OBJECT_TYPES = (pl.Struct,)
TIME_TYPES = (pl.Time,)



from __future__ import annotations

from typing import Optional

from dplib.model import Model


class GithubLicense(Model):
    spdx_id: Optional[str] = None
    name: Optional[str] = None
    key: Optional[str] = None
    html_url: Optional[str] = None



from typing import Any, Union
from urllib.parse import quote

from rdflib import Literal, URIRef


# https://github.com/ckan/ckanext-dcat/blob/master/ckanext/dcat/profiles.py
def create_node(value: Any) -> Union[URIRef, Literal]:
    try:
        stripped_value = value.strip()
        if stripped_value.startswith("http://") or stripped_value.startswith("https://"):
            # only encode this limited subset of characters to avoid more complex URL parsing
            # (e.g. valid ? in query string vs. ? as value).
            # can be applied multiple times, as encoded %xy is left untouched. Therefore, no
            # unquote is necessary beforehand.
            quotechars = " !\"$'()*,;<>[]{|}\\^`"
            for c in quotechars:
                value = value.replace(c, quote(c))
            # although all invalid chars checked by rdflib should have been quoted, try to serialize
            # the object. If it breaks, use Literal instead.
            value = URIRef(value)
            value.n3()
            # URI is fine, return the object
            return value
        else:
            return Literal(value)
    except Exception:
        # In case something goes wrong: use Literal
        return Literal(value)



from __future__ import annotations

from typing import List, Union

from ... import types
from ...errors.metadata import MetadataError
from ...helpers.data import read_data
from ...helpers.path import is_remote_path
from ...helpers.profile import check_metadata_against_jsonschema, read_profile
from ...models import Profile


def check_metadata(
    metadata: Union[str, types.IDict], *, type: str
) -> List[MetadataError]:
    if isinstance(metadata, str):
        metadata = read_data(metadata)

    # Base profile
    profile = Profile.from_dict(read_profile(metadata_type=type))
    errors = check_metadata_against_jsonschema(metadata, profile.jsonSchema)

    # Custom profile
    custom_profile = metadata.get("profile")
    if custom_profile and is_remote_path(custom_profile):
        custom_profile = Profile.from_path(custom_profile)
        errors += check_metadata_against_jsonschema(metadata, custom_profile.jsonSchema)

    return errors



from __future__ import annotations

from typing import List

import pydantic

from ...model import Model
from .foreignKeyReference import ForeignKeyReference


class ForeignKey(Model):
    fields: List[str] = []
    reference: ForeignKeyReference = pydantic.Field(default_factory=ForeignKeyReference)



from __future__ import annotations

import random
from string import ascii_lowercase as letters
from typing import Any, List, Optional

import sqlalchemy as sa
from sqlalchemy.dialects import mysql as ml
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.dialects import registry
from sqlalchemy.schema import Column

from dplib.model import Model
from dplib.models import Field

from . import settings


class SqlField(Model, arbitrary_types_allowed=True):
    column: Column[Any]

    # Converters

    def to_dp(self) -> Field:
        field = Field(name=self.column.name)

        # Type
        if isinstance(self.column.type, ARRAY_TYPES):
            field.type = "array"
        elif isinstance(self.column.type, BOOLEAN_TYPES):
            field.type = "boolean"
        elif isinstance(self.column.type, DATE_TYPES):
            field.type = "date"
        elif isinstance(self.column.type, DATETIME_TYPES):
            field.type = "datetime"
        elif isinstance(self.column.type, INTEGER_TYPES):
            field.type = "integer"
        elif isinstance(self.column.type, NUMBER_TYPES):
            field.type = "number"
        elif isinstance(self.column.type, OBJECT_TYPES):
            field.type = "object"
        elif isinstance(self.column.type, TEXT_TYPES):
            field.type = "string"
        elif isinstance(self.column.type, TIME_TYPES):
            field.type = "time"

        # Description
        if self.column.comment:
            field.description = self.column.comment

        # Constraints
        if not self.column.nullable:
            field.constraints.required = True
        if isinstance(self.column.type, (sa.CHAR, sa.VARCHAR)):
            if self.column.type.length:
                field.constraints.maxLength = self.column.type.length
        if isinstance(self.column.type, sa.CHAR):
            if self.column.type.length:
                field.constraints.minLength = self.column.type.length
        if isinstance(self.column.type, sa.Enum):
            if self.column.enums:
                field.constraints.enum = self.column.enums

        return field

    @classmethod
    def from_dp(
        cls,
        field: Field,
        *,
        dialect: str = settings.DEFAULT_DIALECT,
        table_name: Optional[str] = None,
    ) -> SqlField:
        Check = sa.CheckConstraint
        checks: List[Check] = []
        comment = field.description
        dialect_obj = registry.load(dialect)()
        nullable = not field.constraints.required
        quoted_name = dialect_obj.identifier_preparer.quote(field.name)

        # Type
        column_type = sa.Text
        if field.type == "any":
            column_type = sa.Text
        elif field.type == "boolean":
            column_type = sa.Boolean
        elif field.type == "date":
            column_type = sa.Date
        elif field.type == "datetime":
            column_type = sa.DateTime
        elif field.type == "integer":
            column_type = sa.Integer
        elif field.type == "number":
            column_type = sa.Float
        elif field.type == "string":
            column_type = sa.Text
        elif field.type == "time":
            column_type = sa.Time
        elif field.type == "year":
            column_type = sa.Integer
        if dialect_obj.name == "postgresql":
            if field.type == "array":
                column_type = pg.JSONB
            elif field.type == "geojson":
                column_type = pg.JSONB
            elif field.type == "object":
                column_type = pg.JSONB
            elif field.type == "number":
                column_type = sa.Numeric

        # Unique contstraint
        unique = field.constraints.unique
        if dialect_obj.name == "mysql":
            # MySQL requires keys to have an explicit maximum length
            # https://stackoverflow.com/questions/1827063/mysql-error-key-specification-without-a-key-length
            unique = unique and column_type is not sa.Text

        # Length contraints
        if field.type == "string":
            min = field.constraints.minLength
            max = field.constraints.maxLength
            if min is not None and max is not None and min == max:
                column_type = sa.CHAR(max)
            if max is not None:
                if column_type is sa.Text:
                    column_type = sa.VARCHAR(length=max)
                if dialect_obj.name == "sqlite":
                    checks.append(Check("LENGTH(%s) <= %s" % (quoted_name, max)))
            if min is not None:
                if not isinstance(column_type, sa.CHAR) or dialect_obj.name == "sqlite":
                    checks.append(Check("LENGTH(%s) >= %s" % (quoted_name, min)))

        # Limit contstraints
        if field.type in ["integer", "number"]:
            min = field.constraints.minimum
            max = field.constraints.maximum
            if min is not None:
                checks.append(Check("%s >= %s" % (quoted_name, min)))
            if max is not None:
                checks.append(Check("%s <= %s" % (quoted_name, max)))

        # Pattern constraint
        if field.type == "string":
            val = field.constraints.pattern
            if val is not None:
                if dialect_obj.name == "postgresql":
                    checks.append(Check("%s ~ '%s'" % (quoted_name, val)))
                elif dialect_obj.name != "duckdb":
                    check = Check("%s REGEXP '%s'" % (quoted_name, val))
                    checks.append(check)

        # Enum constraint
        if field.type == "string":
            val = field.constraints.enum
            if val is not None:
                # NOTE: https://github.com/frictionlessdata/frictionless-py/issues/778
                if not table_name:
                    table_name = "".join(random.choice(letters) for _ in range(8))
                enum_name = "%s_%s_enum" % (table_name, field.name)
                quoted_enum_name = dialect_obj.identifier_preparer.quote(enum_name)
                column_type = sa.Enum(*val, name=quoted_enum_name)

        # TODO: shall it use "autoincrement=False"
        # https://github.com/Mause/duckdb_engine/issues/595#issuecomment-1495408566
        column_args = [field.name, column_type] + checks
        column_kwargs = {"nullable": nullable, "unique": unique, "comment": comment}
        column = sa.Column(*column_args, **column_kwargs)  # type: ignore
        return SqlField(column=column)


ARRAY_TYPES = (pg.ARRAY,)
BOOLEAN_TYPES = (sa.Boolean,)
DATE_TYPES = (sa.Date,)
DATETIME_TYPES = (sa.DateTime,)
INTEGER_TYPES = (sa.Integer,)
NUMBER_TYPES = (sa.Float, sa.Numeric)  # type: ignore
STRING_TYPES = (ml.BIT, ml.VARBINARY, ml.VARCHAR, pg.UUID, sa.Text, sa.VARCHAR)  # type: ignore
OBJECT_TYPES = (pg.JSONB, pg.JSON)
TEXT_TYPES = (sa.Text,)
TIME_TYPES = (sa.Time,)



from __future__ import annotations

from typing import Optional

from dplib.model import Model


class DataciteSubject(Model):
    subject: str
    subjectScheme: Optional[str] = None
    schemeUri: Optional[str] = None
    lang: Optional[str] = None



from __future__ import annotations

from typing import List, Optional

from dplib.model import Model

from .creator import ZenodoCreator
from .subject import ZenodoSubject


class ZenodoMetadata(Model):
    creators: List[ZenodoCreator] = []
    description: Optional[str] = None
    publication_date: Optional[str] = None
    publisher: Optional[str] = None
    subjects: List[ZenodoSubject] = []
    title: Optional[str] = None
    version: Optional[str] = None



from __future__ import annotations

from typing import List, Optional

from dplib.model import Model
from dplib.models import Contributor, License, Package

from .organization import CkanOrganization
from .resource import CkanResource
from .tag import CkanTag

# References:
# - https://demo.ckan.org/api/3/action/package_show?id=sample-dataset-1


class CkanPackage(Model):
    resources: List[CkanResource] = []

    organization: Optional[CkanOrganization] = None
    tags: List[CkanTag] = []

    id: Optional[str] = None
    name: Optional[str] = None
    title: Optional[str] = None
    notes: Optional[str] = None
    version: Optional[str] = None
    notes: Optional[str] = None
    license_id: Optional[str] = None
    license_title: Optional[str] = None
    license_url: Optional[str] = None
    author: Optional[str] = None
    author_email: Optional[str] = None
    maintainer: Optional[str] = None
    maintainer_email: Optional[str] = None
    metadata_created: Optional[str] = None
    metadata_modified: Optional[str] = None

    # Converters

    def to_dp(self):
        package = Package()

        # Name
        if self.name:
            package.name = self.name

        # Title
        if self.title:
            package.title = self.title

        # Description
        if self.notes:
            package.description = self.notes

        # Version
        if self.version:
            package.version = self.version

        # Created
        if self.metadata_created:
            package.created = self.metadata_created

        # License
        if self.license_id:
            license = License(name=self.license_id)
            if self.license_title:
                license.title = self.license_title
            if self.license_url:
                license.url = self.license_url
            package.licenses.append(license)

        # Contributors
        if self.author:
            contributor = Contributor(title=self.author, role="author")
            if self.author_email:
                contributor.email = self.author_email
            package.contributors.append(contributor)
        if self.maintainer:
            contributor = Contributor(title=self.maintainer, role="maintainer")
            if self.maintainer_email:
                contributor.email = self.maintainer_email
            package.contributors.append(contributor)

        # Resources
        for item in self.resources:
            resource = item.to_dp()
            package.resources.append(resource)

        # Keywords
        for tag in self.tags:
            package.keywords.append(tag.name)

        # Custom
        if self.id:
            package.custom["ckan:id"] = self.id

        return package

    @classmethod
    def from_dp(cls, package: Package) -> CkanPackage:
        ckan = CkanPackage()

        # Name
        if package.name:
            ckan.name = package.name

        # Title
        if package.title:
            ckan.title = package.title

        # Description
        if package.description:
            ckan.notes = package.description

        # Version
        if package.version:
            ckan.version = package.version

        # License
        if package.licenses:
            license = package.licenses[0]
            ckan.license_id = license.name
            if license.title:
                ckan.license_title = license.title
            if license.path:
                ckan.license_url = license.path

        # Contributors
        for contributor in package.contributors:
            if contributor.role == "author":
                ckan.author = contributor.title
                ckan.author_email = contributor.email
            elif contributor.role == "maintainer":
                ckan.maintainer = contributor.title
                ckan.maintainer_email = contributor.email

        # Resources
        for resource in package.resources:
            item = CkanResource.from_dp(resource)
            if item:
                ckan.resources.append(item)

        # Keywords
        for keyword in package.keywords:
            tag = CkanTag(name=keyword)
            ckan.tags.append(tag)

        return ckan



from typing import Optional

from ..model import Model


class Contributor(Model):
    title: Optional[str] = None
    path: Optional[str] = None
    email: Optional[str] = None
    role: Optional[str] = None
    organization: Optional[str] = None



from __future__ import annotations

from dplib.actions.dialect.check import check_dialect

from ...helpers.check import print_check_results
from ...options.path import Path
from .main import program


@program.command(name="check")
def command(
    path: str = Path,
):
    """Check the validity of a Table Dialect descriptor."""
    errors = check_dialect(path)
    print_check_results(path, errors)



from __future__ import annotations

from typing import List, Optional

import pydantic

from ...model import Model
from ..contributor import Contributor
from ..license import License
from ..profile import Profile
from ..resource import Resource
from ..source import Source


class Package(Model):
    basepath: Optional[str] = pydantic.Field(default=None, exclude=True)

    id: Optional[str] = None
    name: Optional[str] = None
    profile: Optional[str] = None
    resources: List[Resource] = []

    title: Optional[str] = None
    description: Optional[str] = None
    homepage: Optional[str] = None
    version: Optional[str] = None
    licenses: List[License] = []
    sources: List[Source] = []
    contributors: List[Contributor] = []
    keywords: List[str] = []
    image: Optional[str] = None
    created: Optional[str] = None

    def model_post_init(self, _):
        for resource in self.resources:
            resource.basepath = self.basepath

    # Getters

    def get_profile(self) -> Optional[Profile]:
        if self.profile:
            return Profile.from_path(self.profile)

    def get_resource(
        self, *, name: Optional[str] = None, path: Optional[str] = None
    ) -> Optional[Resource]:
        for resource in self.resources:
            if name and resource.name == name:
                return resource
            if path and resource.path == path:
                return resource

    # Setters

    def add_resource(self, resource: Resource) -> None:
        resource.basepath = self.basepath
        self.resources.append(resource)

    # Methods

    def dereference(self):
        for resource in self.resources:
            resource.dereference()



from __future__ import annotations

from typing import Optional

from dplib.helpers.resource import slugify_name
from dplib.model import Model
from dplib.models import Resource


class CkanResource(Model):
    name: str
    created: Optional[str] = None
    description: Optional[str] = None
    format: Optional[str] = None  # NOTE: uppercased
    hash: Optional[str] = None
    id: Optional[str] = None
    last_modified: Optional[str] = None
    metadata_modified: Optional[str] = None
    mimetype: Optional[str] = None
    size: Optional[int] = None

    # Converters

    def to_dp(self) -> Resource:
        resource = Resource(path=self.name, name=slugify_name(self.name))

        # Format
        if self.format:
            resource.format = self.format.lower()

        # Mediatype
        if self.mimetype:
            resource.mediatype = self.mimetype

        # Size
        if self.size:
            resource.bytes = self.size

        # Custom
        if self.id:
            resource.custom["ckan:id"] = self.id

        return resource

    @classmethod
    def from_dp(cls, resource: Resource) -> Optional[CkanResource]:
        if not resource.path or not isinstance(resource.path, str):
            return

        # Path
        ckan = CkanResource(name=resource.path)

        # Format
        if resource.format:
            ckan.format = resource.format.upper()

        # Mediatype
        if resource.mediatype:
            ckan.mimetype = resource.mediatype

        # Bytes
        if resource.bytes:
            ckan.size = resource.bytes

        return ckan



from dplib.models import Profile


def test_profile_from_dict():
    data = {"metadataProfile": {"name": "name"}, "additionalProperties": False}
    profile = Profile.from_dict(data)
    assert profile.name == "name"


def test_profile_to_dict():
    profile = Profile()
    profile.name = "name"
    profile.jsonSchema["additionalProperties"] = False
    assert profile.to_dict() == {
        "metadataProfile": {"name": "name"},
        "additionalProperties": False,
    }
